"""
Parse Kaycha Labs COAs
Copyright (c) 2022-2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
Created: 9/17/2022
Updated: 6/3/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse Kaycha Labs COA PDFs.

Data Points:

    ✓ lab_id
    ✓ product_name
    ✓ product_type
    ✓ batch_number
    ✓ product_size
    - serving_size
    - servings_per_package
    ✓ sample_weight
    ✓ sample_id (augmented)
    ✓ strain_name
    ✓ lab
    ✓ lab_image_url
    - lab_license_number
    ✓ lab_address
    ✓ lab_street
    ✓ lab_city
    ✓ lab_county (augmented)
    ✓ lab_state
    ✓ lab_zipcode
    ✓ lab_phone
    ✓ lab_email
    ✓ lab_website
    ✓ lab_latitude (augmented)
    ✓ lab_longitude (augmented)
    ✓ producer
    - producer_address
    - producer_street
    - producer_city
    - producer_state
    - producer_zipcode
    - producer_license_number
    ✓ distributor
    ✓ distributor_address
    ✓ distributor_street
    ✓ distributor_city
    ✓ distributor_state
    ✓ distributor_zipcode
    - distributor_license_number
    ✓ date_collected
    ✓ date_tested
    ✓ date_received
    - images
    ✓ analyses
    ✓ {analysis}_status
    ✓ status
    ✓ lab_results_url
    ✓ coa_urls
    ✓ methods
    ✓ total_cannabinoids
    ✓ total_thc
    ✓ total_cbd
    ✓ total_terpenes
    ✓ results
        ✓ cannabinoids
        ✓ terpenes
        ✓ pesticides
        ✓ heavy_metals
        ✓ microbes
        ✓ mycotoxins
        ✓ residual_solvents
        ✓ foreign_matter
        ✓ water_activity
        ✓ moisture
"""
# Standard imports.
from ast import literal_eval
import base64
from datetime import datetime
import json
import io
import re
import os
import tempfile
from typing import Any, Optional

# External imports.
import pandas as pd
import pdfplumber
from PIL import Image

# Internal imports.
from cannlytics import firebase
from cannlytics import __version__
from cannlytics.data.data import (
    create_hash,
    create_sample_id,
    find_first_value,
)
from cannlytics.utils.constants import ANALYTES
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    split_list,
    strip_whitespace,
)

# It is assumed that the lab has the following details.
KAYCHA_LABS = {
    'coa_algorithm': 'kaycha.py',
    'coa_algorithm_entry_point': 'parse_kaycha_coa',
    'lims': 'Kaycha Labs',
    'lab': 'Kaycha Labs',
    'lab_image_url': 'https://www.kaychalabs.com/wp-content/uploads/2020/06/newlogo-2.png',
    'lab_address': '4101 SW 47th Ave, Suite 105, Davie, FL 33314',
    'lab_street': '4101 SW 47th Ave, Suite 105',
    'lab_city': 'Davie',
    'lab_county': 'Broward',
    'lab_state': 'FL',
    'lab_zipcode': '33314',
    'lab_phone': '833-465-8378',
    'lab_email': 'info@kaychalabs.com',
    'lab_website': 'https://www.kaychalabs.com/',
    'lab_latitude': 26.071350,
    'lab_longitude': -80.210750,
    # FIXME: Make license number dynamic as Kaycha Labs operate in multiple states.
    'lab_license_number': 'CMTL-0002',
}
KAYCHA_LABS_COA = {
    'fields': {
        'Matrix': 'product_type',
        'Sample': 'lab_id',
        'Harvest/Lot ID': 'source_id',
        'Batch#': 'batch_number',
        'Cultivation Facility': 'producer',
        'Processing Facility': 'processor',
        'Distributor Facility': 'distributor',
        'Source Facility': 'source',
        'Seed to Sale': 'traceability_id',
        'Batch Date': 'date_harvested', # FIXME: This date is not being parsed correctly.
        'Sample Size Received': 'sample_weight',
        'Total Batch Size': 'batch_size',
        'Retail Product Size': 'product_size',
        'Ordered': 'date_received',
        'Sampled': 'date_collected',
        'Completed': 'date_tested',
        # 'Sampling Method': 'method_sampling',
    },
    'screening_analyses': [
        'pesticides',
        'heavy_metals',
        'microbes',
        'mycotoxins',
        'residuals_solvents',
        'foreign_matter',
        'water_activity',
        'moisture',
        'terpenes',
    ],
    'screening_statuses': {
        'PASSED': 'Pass',
        'FAILED': 'Fail',
        'TESTED': 'Tested',
        'NOT TESTED': 'N/A',
    },
}


def get_kaycha_terpenes(parser, page, obs, results):
    """Get terpenes from a Kaycha Labs COA PDF."""
    # Check if the page has terpenes.
    if 'Terpenes TESTED' not in page.extract_text():
        return obs, results

    # Split the page in half.
    left = page.within_bbox((0, 0, page.width * 0.5, page.height)).extract_text()
    right = page.within_bbox((page.width * 0.5, 0, page.width, page.height)).extract_text()

    # Get the relevant portions.
    left = left.split('TOTAL TERPENES')[-1].split('This Kaycha Labs Certification shall not be reproduced')[0]
    right = right.split('(%)')[-1].split('Analyzed by')[0]
    left_lines = [x for x in left.split('\n') if x]
    right_lines = [x for x in right.split('\n') if x]

    # Get total terpenes.
    total = left_lines[-1].split('(%)')[-1].strip()
    obs['total_terpenes'] = convert_to_numeric(total, strip=True)

    # Get individual terpenes.
    lines = left_lines[1:-1] + right_lines
    for line in lines:
        first_value = find_first_value(line)
        name = line[:first_value].strip()
        key = parser.analytes.get(snake_case(name), snake_case(name))
        values = line[first_value:].strip().split(' ')
        results.append({
            'analysis': 'terpenes',
            'key': key,
            'lod': convert_to_numeric(values[0]),
            'name': name,
            'units': 'percent',
            'value': convert_to_numeric(values[-1]),
            'mg_unit': convert_to_numeric(values[1]),
        })

    # Return the observation and results.
    return obs, results


# UNDER DEVELOPMENT:
def parse_kaycha_coa(
        parser,
        doc: Any,
        temp_path: Optional[str] = None,
        coa_parameters: Optional[dict] = KAYCHA_LABS_COA,
        **kwargs,
    ) -> dict:
    """Parse a Kaycha Labs COA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """

    # FIXME: If the `doc` is a URL, then download the PDF to `temp_path`.
    # And then use the downloaded PDF as the doc.
    if temp_path is None:
        temp_path = tempfile.gettempdir()

    # Get the lab's parameters.
    screening_analyses = coa_parameters['screening_analyses']
    screening_statuses = coa_parameters['screening_statuses']

    # Read the PDF.
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.replace('\\', '/').split('/')[-1]

    # Get the text of the first page.
    front_page = report.pages[0]
    text = front_page.extract_text()
    text = text.split('This Kaycha Labs Certification shall not be reproduced')[0]
    lines = text.split('\n')

    # Get the lab results URL from the QR code.
    coa_url = parser.find_pdf_qr_code_url(front_page)
    obs['lab_results_url'] = coa_url

    # Format `coa_urls`
    if coa_url is not None:
        filename = coa_url.split('/')[-1].split('?')[0] + '.pdf'
        obs['coa_urls'] = json.dumps([{'url': coa_url, 'filename': filename}])

    # Get lab details.
    parts = lines[5].split(',')
    city, state, zipcode = [x.strip() for x in parts[:3]]
    obs['lab_street'] = lines[4].title()
    obs['lab_city'] = city.title()
    obs['lab_state'] = state
    obs['lab_zipcode'] = zipcode

    # Get sample details.
    obs['product_name'] = lines[1]
    obs['strain_name'] = lines[2]
    obs['product_type'] = lines[3].split(':')[-1].strip()
    obs['lab_id'] = lines[6].split(':')[-1].strip()

    # Get additional sample details.
    # FIXME: These are dry weight! People want wet weight.
    results = []
    totals = ['total_thc', 'total_cbd', 'total_cannabinoids']
    for i, line in enumerate(lines):

        # Get sample fields.
        for key, value in coa_parameters['fields'].items():
            field = key.lower()
            cell = line.lower()
            if f'{field}:' in cell or f'{field} :' in cell:
                obs[value] = line.split(':')[-1].strip()
            elif f'{field}#' in cell:
                obs[value] = line.split('#')[-1].strip()

        # Get distributor, distributor address, and testing status.
        if '|' in line:
            name = line.split('|')[-1]
            for k, v in screening_statuses.items():
                if k in line:
                    name = name.replace(k, '').strip()
                    obs['status'] = v
            street = lines[i + 1]
            city, state, zipcode, _ = [x.strip() for x in lines[i + 2].split(',')]
            obs['distributor'] = name
            obs['distributor_address'] = f'{street}, {city}, {state} {zipcode}'
            obs['distributor_street'] = street
            obs['distributor_city'] = city
            obs['distributor_state'] = state
            obs['distributor_zipcode'] = zipcode
        
        # Get totals.
        if 'Total THC' in line and obs.get('total_thc') is None:
            values = [convert_to_numeric(x, strip=True) for x in lines[i + 1].split(' ')]
            for v, value in enumerate(values):
                obs[totals[v]] = value
        
        # Get cannabinoids.
        # Optional: Also record mg/unit, method, and instrument.
        if line.startswith('%'):
            analytes = lines[i - 1].split(' (DRY)')[0].split(' ')
            if len(analytes) <= 2:
                analytes = lines[i - 2].split(' ')
            keys = [parser.analytes.get(snake_case(x), snake_case(x)) for x in analytes]
            values = [convert_to_numeric(x) for x in line.split('Analyte')[0].lstrip('% ').split(' ') if x != '']
            if lines[i + 2].startswith('Analysis'):
                lod = [convert_to_numeric(x) for x in lines[i + 3].lstrip('LOD ').split(' ') if x != '']
            else:
                lod = [convert_to_numeric(x) for x in lines[i + 2].lstrip('LOD ').split(' ') if x != '']
            for k, key in enumerate(keys):
                results.append({
                    'analysis': 'cannabinoids',
                    'key': key,
                    'name': analytes[k],
                    'value': values[k],
                    'unit': 'percent',
                    'lod': lod[k],
                })
            break

    # FIXME: Get lab license number.
    # State License # CMTL-0002

    # TODO: Try to get producer details from licenses data.
    # - producer_address
    # - producer_street
    # - producer_city
    # - producer_state
    # - producer_zipcode
    # - producer_license_number

    # Get analyses and status data.
    analyses = []
    safety_results = text.split('SAFETY RESULTS')[-1].split('Cannabinoid')[0].split('\n')
    statuses = safety_results[2].replace('NOT TESTED', 'NOT_TESTED').split(' ')
    for i, status in enumerate(statuses):
        if status == 'NOT_TESTED':
            continue
        if status == 'Solvents':
            status = safety_results[3]
        analysis = screening_analyses[i]
        obs[f'{analysis}_status'] = screening_statuses.get(status, status)
        analyses.append(analysis)

    # Get terpene results.
    # Optional Add "Analysis Method" and "Instrument Used".
    num_of_pages = len(report.pages)
    if num_of_pages >= 2:
        page = report.pages[1]
        obs, results = get_kaycha_terpenes(parser, page, obs, results)

    # Get additional analyses for full panel COAs.
    rows = []
    if num_of_pages > 2:
        for page in report.pages[2:]:

            # Handle residual solvents page.
            page_text = page.extract_text()
            if 'Residual Solvents' in page_text:
                lines = page_text.split('Residual Solvents')[1].strip().split('\n')
                for line in lines[2:]:
                    first_value = find_first_value(line)
                    name = line[:first_value].strip()
                    key = parser.analytes.get(snake_case(name), snake_case(name))
                    values = line[first_value:].strip().split(' ')
                    results.append({
                        'analysis': 'residual_solvents',
                        'key': key,
                        'lod': convert_to_numeric(values[0]),
                        'name': name,
                        'units': values[1],
                        'value': convert_to_numeric(values[-1]),
                        'status': values[-2],
                        'limit': convert_to_numeric(values[2]),
                    })
                continue

            # Split the page in half.
            left = page.within_bbox((0, 0, page.width * 0.49, page.height)).extract_text()
            right = page.within_bbox((page.width * 0.49, 0, page.width, page.height)).extract_text()

            # Get the relevant portions.
            left = re.split(r'Page \d+ of \d+', left)[-1].split('This Kaycha Labs Certification shall not be reproduced')[0]
            left_lines = [x for x in left.split('\n') if x]
            right_lines = [x for x in right.split('\n') if x]

            # Remove unnecessary rows.
            left_lines = [x for x in left_lines if ':' not in x]
            right_lines = [x for x in right_lines if ':' not in x]
            rows.extend(left_lines + right_lines)

    # Keep only rows with values.
    rows = [x for x in rows if len(x.split(' ')) > 3]
    rows = [x for x in rows if x != obs['product_name']]
    rows = [x for x in rows if ', US' not in x]
    cells = [x for x in rows if x[:3].isupper() and not x.startswith('SOP.')]
    cells = [x .replace('*', '').replace('Not Present', 'ND') for x in cells]

    # Define contaminants.
    microbes = ['ASPERGILLUS', 'SALMONELLA', 'ESCHERICHIA', 'TOTAL YEAST']
    mycotoxins = ['AFLATOXIN', 'OCHRATOXIN']
    heavy_metals = ['TOTAL CONTAMINANT LOAD METALS', 'ARSENIC', 'CADMIUM', 'MERCURY', 'LEAD']
    contaminants = [{'name': microbe, 'analysis': 'microbes', 'units': 'CFU/g'} for microbe in microbes] + \
        [{'name': mycotoxin, 'analysis': 'mycotoxins', 'units': 'ppm'} for mycotoxin in mycotoxins] + \
        [{'name': metal, 'analysis': 'heavy_metals', 'units': 'ppm'} for metal in heavy_metals]

    # Get contaminants: microbes, mycotoxins, and heavy metals.
    remove = []
    for contaminant in contaminants:
        analyte = contaminant['name']
        analysis = contaminant['analysis']
        units = contaminant['units']
        for cell in cells:
            if analyte in cell:
                remove.append(cell)
                line = cell
                line = line.replace('TOTAL YEAST AND MOLD', 'MOLD')
                first_value = find_first_value(line)
                name = line[:first_value].strip()
                key = parser.analytes.get(snake_case(name), snake_case(name))
                values = line[first_value:].strip().split(' ')
                if len(values) == 2:
                    results.append({
                        'analysis': analysis,
                        'key': key,
                        'lod': None,
                        'name': name,
                        'units': units,
                        'value': convert_to_numeric(values[0]),
                        'status': values[-1],
                    })
                else:
                    results.append({
                        'analysis': analysis,
                        'key': key,
                        'lod': convert_to_numeric(values[0]),
                        'name': name,
                        'units': values[1],
                        'value': convert_to_numeric(values[2]),
                        'status': values[-2],
                        'limit': convert_to_numeric(values[-1]),
                    })

    # Remove collected cells.
    cells = [x for x in cells if x not in remove]

    # Get any pesticides analyzed.
    if cells:
        for line in cells:
            first_value = find_first_value(line)
            name = line[:first_value].strip()
            key = parser.analytes.get(snake_case(name), snake_case(name))
            values = line[first_value:].strip().split(' ')
            results.append({
                'analysis': 'pesticides',
                'key': key,
                'lod': convert_to_numeric(values[0]),
                'name': name,
                'units': values[1],
                'value': convert_to_numeric(values[-1]),
                'status': values[-2],
                'limit': convert_to_numeric(values[2]),
            })

    # Get moisture content, water activity, and foreign matter.
    screens = [
        {'name': 'Filth and Foreign Material', 'key': 'foreign_matter', 'units': 'percent'},
        {'name': 'Moisture Content', 'key': 'moisture_content', 'units': 'percent'},
        {'name': 'Water Activity', 'key': 'water_activity', 'units': 'aW'},
    ]
    for screen in screens:
        name = screen['name']
        for row in rows:
            if row.startswith(name):
                values = row.replace(name, '').strip().split(' ')
                results.append({
                    'name': name,
                    'key': screen['key'],
                    'value': convert_to_numeric(values[2]),
                    'units': screen['units'],
                    'lod': convert_to_numeric(values[0]),
                    'limit': convert_to_numeric(values[-1]),
                    'status': values[-2]
                })
                break

    # FIXME: Save the image data to Firebase Storage.
    # image_index = 5
    # try:
    #     temp_dir = tempfile.gettempdir()
    #     file_ref = f'data/lab_results/images/{lab_id}/image_data.png'
    #     file_path = os.path.join(temp_dir, 'image_data.png')
    #     image_data = parser.get_pdf_image_data(front_page, image_index=image_index)
    #     parser.save_image_data(image_data, image_file=file_path)
    #     bucket_name = config['FIREBASE_STORAGE_BUCKET']
    #     firebase.upload_file(file_ref, file_path, bucket_name=bucket_name)
    #     download_url = firebase.get_file_url(file_ref, bucket_name=bucket_name)
    #     obs['images'] = [{'ref': file_ref, 'url': download_url, 'filename': 'image_data.png'}]
    # except:
    #     print('Failed to get image data.')
    #     obs['images'] = []

    # Get all the lines with methods.
    method_lines = ''
    for page in report.pages:
        text = page.extract_text()
        lines = text.split('\n')
        for line in lines:
            if 'Analysis Method :' in line:
                method_lines += line + '\n'

    # Find all matches in the text and return them.
    pattern = r'SOP\.[\w\.\-\(\)]+'
    methods = re.findall(pattern, method_lines)

    # Close the report.
    report.close()

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Finish data collection with a freshly minted sample ID.
    obs = {**KAYCHA_LABS, **obs}
    obs['analyses'] = json.dumps(list(set(analyses)))
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['methods'] = json.dumps(methods)
    obs['results'] = json.dumps(results)
    obs['results_hash'] = create_hash(results)
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs.get('producer', obs.get('date_tested', 'cannlytics.eth')),
    )
    obs['sample_hash'] = create_hash(obs)
    return obs


# === Tests ===
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc

    # [✓] TEST: Identify LIMS.
    parser = CoADoc()
    docs = [
        '../../../../tests/assets/coas/kaycha-labs/DA30318004-001-Original.pdf',
    ]
    for doc in docs:
        lims = parser.identify_lims(doc, lims={'Kaycha Labs': KAYCHA_LABS})
        assert lims == 'Kaycha Labs'

    # [ ] TEST: Parse a full panel COA PDF.
    doc = '../../../../tests/assets/coas/kaycha-labs/DA30318004-001-Original.pdf'
    data = parse_kaycha_coa(parser, doc)
    assert data is not None


    # [ ] TEST: Parse a cannabinoid and terpene only COA PDF.
    # doc = 'D://data/florida/lab_results/.datasets/pdfs/MMTC-2019-0015/GA11104001-001.pdf'


    # [ ] TEST: Parse a COA with residual solvent screening.
    doc = '../../../../tests/assets/coas/kaycha-labs/DA20330006-008.pdf'


    # [ ] TEST: Parse a non-mandatory COA PDF.
    # doc = 'D://data/florida/lab_results/.datasets/pdfs/MMTC-2019-0015/GA11104001-001.pdf'


    # [ ] TEST: Parse Kaycha Labs COAs from URL.
    # urls = [
    #     'https://tn.yourcoa.com/api/coa-download?sample=KN20119003-002&wl_id=291',
    # ]


    # [ ] TEST: Parse a full panel COA from a URL.
    # urls = [
    #     'https://getfluent.com/wp-content/uploads/2023/03/DA30318004-001-Original.pdf',
    # ]


    # [ ] TEST: Parse COAs for various product types.


    # [ ] TEST: Parse historic COAs.
    # doc = '../../../../tests/assets/coas/kaycha-labs/DA90911001-009.pdf'


    # TODO: Handle Evio Labs COAs.



"""
Firebase Interface Initialization | Cannlytics
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 5/5/2022
Updated: 1/10/2023
"""
from .firebase import (
    MAX_BATCH_SIZE,
    access_secret_version,
    add_secret_version,
    add_to_array,
    create_custom_claims,
    create_custom_token,
    create_document,
    create_id,
    create_id_from_datetime,
    create_log,
    create_reference,
    create_secret,
    create_session_cookie,
    create_short_url,
    create_user,
    delete_collection,
    delete_document,
    delete_field,
    delete_file,
    delete_user,
    download_file,
    download_files,
    export_data,
    generate_password_reset_link,
    get_collection,
    get_custom_claims,
    get_document,
    get_file_url,
    get_id_timestamp,
    get_random_string,
    get_user,
    get_users,
    import_data,
    increment_value,
    initialize_firebase,
    list_files,
    remove_from_array,
    rename_file,
    revoke_refresh_tokens,
    update_custom_claims,
    update_document,
    update_documents,
    update_user,
    upload_file,
    upload_files,
    verify_session_cookie,
    verify_token,
)

__all__ = [
    MAX_BATCH_SIZE,
    access_secret_version,
    add_secret_version,
    add_to_array,
    create_custom_claims,
    create_custom_token,
    create_document,
    create_id,
    create_id_from_datetime,
    create_log,
    create_reference,
    create_secret,
    create_session_cookie,
    create_short_url,
    create_user,
    delete_collection,
    delete_document,
    delete_field,
    delete_file,
    delete_user,
    download_file,
    download_files,
    export_data,
    generate_password_reset_link,
    get_collection,
    get_custom_claims,
    get_document,
    get_file_url,
    get_id_timestamp,
    get_random_string,
    get_user,
    get_users,
    import_data,
    increment_value,
    initialize_firebase,
    list_files,
    remove_from_array,
    rename_file,
    revoke_refresh_tokens,
    update_custom_claims,
    update_document,
    update_documents,
    update_user,
    upload_file,
    upload_files,
    verify_session_cookie,
    verify_token,
]



"""
CoADoc | A Certificate of Analysis (COA) Parser
Copyright (c) 2022-2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 7/15/2022
Updated: 6/11/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Certificates of analysis (CoAs) are abundant for cannabis cultivators,
    processors, retailers, and consumers too, but the data is often
    locked away. Rich, valuable laboratory data so close, yet so far
    away! CoADoc puts these vital data points in your hands by
    parsing PDFs and URLs, finding all the data, standardizing the data,
    and cleanly returning the data to you.

Future work:

    - [ ] Integrate `create_hash` into `save` and `standardize`.
    - [ ] Improve the `standardize` method.

"""
# Standard imports.
import ast
import base64
from datetime import datetime
import importlib
from io import BytesIO
import json
import operator
import os
from typing import Any, List, Optional

# External imports.
import cv2
import numpy as np
try:
    import openai
except ImportError:
    print('Unable to find `openai` package. This tool is used for parsing with AI.')
import openpyxl
import pandas as pd
import requests
import pdfplumber
from PIL import Image
from pypdf import PdfMerger
try:
    from pyzbar import pyzbar
    # from pyzbar.pyzbar import decode
except ImportError:
    print('Unable to find `zbar` library. This tool is used for decoding QR codes.')
try:
    from pytesseract import image_to_pdf_or_hocr
except ImportError:
    print('Unable to find `Tesseract` library. This tool is used for OCR.')
try:
    from wand.image import Image as magick_wand
    from wand.color import Color
except ImportError:
    print('Unable to find `ImageMagick` library. This tool is used for OCR.')

# Internal imports.
from cannlytics.data.data import create_hash, write_to_worksheet
from cannlytics.data.web import download_google_drive_file
from cannlytics.utils import (
    convert_to_numeric,
    dump_column,
    get_directory_files,
    sandwich_list,
    reorder_columns,
    snake_case,
    unzip_files,
)
from cannlytics.utils.constants import (
    ANALYSES,
    ANALYTES,
    CODINGS,
    DEFAULT_HEADERS,
    STANDARD_FIELDS,
)

# Lab and LIMS CoA parsing algorithms.
from cannlytics.data.coas.coa_ai import parse_coa_with_ai
from cannlytics.data.coas.algorithms.acs import ACS_LABS
from cannlytics.data.coas.algorithms.anresco import ANRESCO
from cannlytics.data.coas.algorithms.cannalysis import CANNALYSIS
from cannlytics.data.coas.algorithms.confidence import CONFIDENCE
from cannlytics.data.coas.algorithms.confidentcannabis import CONFIDENT_CANNABIS
from cannlytics.data.coas.algorithms.greenleaflab import GREEN_LEAF_LAB
from cannlytics.data.coas.algorithms.kaycha import KAYCHA_LABS
from cannlytics.data.coas.algorithms.mcrlabs import MCR_LABS
from cannlytics.data.coas.algorithms.sclabs import SC_LABS
from cannlytics.data.coas.algorithms.sonoma import SONOMA
from cannlytics.data.coas.algorithms.tagleaf import TAGLEAF
from cannlytics.data.coas.algorithms.steephill import STEEPHILL
from cannlytics.data.coas.algorithms.veda import VEDA_SCIENTIFIC

# Labs and LIMS that CoADoc can parse.
LIMS = {
    'ACS Labs': ACS_LABS,
    'Anresco Laboratories': ANRESCO,
    'Cannalysis': CANNALYSIS,
    'Confidence Analytics': CONFIDENCE,
    'Confident Cannabis': CONFIDENT_CANNABIS,
    'Green Leaf Lab': GREEN_LEAF_LAB,
    'Kaycha Labs': KAYCHA_LABS,
    'MCR Labs': MCR_LABS,
    'SC Labs': SC_LABS,
    'Sonoma Lab Works': SONOMA,
    'TagLeaf LIMS': TAGLEAF,
    'Steep Hill': STEEPHILL,
    'Veda Scientific': VEDA_SCIENTIFIC,
}

# Default preferred order for DataFrame columns.
# TODO: Add `sample_hash` and `results_hash` to the beginning.
DEFAULT_COLUMN_ORDER = ['product_name', 'producer',
    'product_type', 'date_tested']

# Default nuisance columns to remove during standardization.
DEFAULT_NUISANCE_COLUMNS = ['received_by', 'sampled_by', 'Unnamed: 1',
    'None_method', 'loss_on_drying_moisture', '_3_5_grams',
    'index', 'other_analyses', 'wildcard']

# Default columns to apply codings and be treated as numeric.
DEFAULT_NUMERIC_COLUMNS = ['value', 'mg_g', 'lod', 'loq', 'limit', 'margin_of_error']

# Encountered Metrc prefixes used to identify Metrc IDs.
METRC_PREFIXES = ['1A40603', '1A40A01']

# A custom `ValueError` message to raise when no known LIMS is identified.
UNIDENTIFIED_LIMS = 'COA not recognized as a COA from a known lab: '
UNIDENTIFIED_LIMS += ', '.join([f'"{x}"' for x in LIMS.keys()])


def convert_pdf_to_images(
        filename: str,
        output_path: str,
        resolution: Optional[int] = 300,
    ) -> List[str]:
    """Convert a PDF into images, with images named as `{filename}-{#}.png`.
    The function replaces the images' alpha channels with a white background.
    Args:
        filename (str): The name of a PDF to convert to images.
        output_path (str): A directory where the images should be generated.
        resolution (int): The resolution of the generated images, 300 by
            default (optional.)
    Returns:
        (list): Returns a list of the filenames of the images generated.
    Authors: Thibaut Mattio, Averner <https://stackoverflow.com/a/42525093/5021266>
    License: CC BY-SA 4.0 <https://creativecommons.org/licenses/by-sa/4.0/>
    """
    image_files = []
    os.environ['MAGICK_TMPDIR'] = output_path
    all_pages = magick_wand(filename=filename, resolution=resolution)
    for i, page in enumerate(all_pages.sequence):
        with magick_wand(page) as img:
            img.format = 'png'
            img.background_color = Color('white')
            img.alpha_channel = 'remove'
            image_filename = os.path.basename(filename)
            image_filename = os.path.splitext(image_filename)[0]
            image_filename = '{}-{}.png'.format(image_filename, i)
            image_filename = os.path.join(output_path, image_filename)
            img.save(filename=image_filename)
            image_files.append(image_filename)
    # Help `ImageMagick` remove temporary files.
    for i in os.listdir(output_path):
        magick_path = os.path.join(output_path, i)
        if os.path.isfile(magick_path) and i.startswith('magick-'):
            os.remove(magick_path)
    return image_files


class CoADoc:
    """Parse data from certificate of analysis (CoA) PDFs or URLs."""

    def __init__(
            self,
            lims: Optional[Any] = None,
            analyses: Optional[dict] = None,
            analytes: Optional[dict] = None,
            codings: Optional[dict] = None,
            column_order: Optional[list] = None,
            nuisance_columns: Optional[list] = None,
            numeric_columns: Optional[list] = None,
            standard_fields: Optional[dict] = None,
            google_maps_api_key: Optional[str] = None,
            headers: Optional[dict] = None,
            init_all: Optional[bool] = True,
        ) -> None:
        """Initialize CoA parser.
        Args:
            analyses (dict): A dictionary of analyses,
                standard analyses are used by default.
            analytes (dict): A dictionary of analytes,
                standard analytes are used by default.
            codings (dict): A dictionary of value codings,
                standard codings are used by default.
            standard_fields (dict): A dictionary of field keys,
                standard fields are used by default.
            headers (dict): Headers for HTTP requests,
                standard headers are used by default.
            lims (str or dict): Specific LIMS to parse CoAs,
                all verified LIMS are checked by default,
                until a matching LIMS is found.
            init_all (bool): Initialize all of the parsing routines
                for all verified LIMS.
            google_maps_api_key (str): A Google Maps API key
                if you also want GIS data included.
        """
        # Setup.
        self.driver = None
        self.options = None
        self.session = None
        self.service = None

        # Define analyses.
        self.analyses = analyses
        if analyses is None:
            self.analyses = ANALYSES

        # Define analytes.
        self.analytes = analytes
        if analytes is None:
            self.analytes = ANALYTES

        # Define codings.
        self.codings = codings
        if codings is None:
            self.codings = CODINGS

        # Define default column / field order.
        self.column_order = column_order
        if column_order is None:
            self.column_order = DEFAULT_COLUMN_ORDER

        # Define default nuisance columns / fields to remove.
        self.nuisance_columns = nuisance_columns
        if nuisance_columns is None:
            self.nuisance_columns = DEFAULT_NUISANCE_COLUMNS

        # Define default columns / fields to treat as numeric.
        self.numeric_columns = numeric_columns
        if numeric_columns is None:
            self.numeric_columns = DEFAULT_NUMERIC_COLUMNS 

        # Define fields.
        self.fields = standard_fields
        if standard_fields is None:
            self.fields = STANDARD_FIELDS

        # Define headers.
        self.headers = headers
        if headers is None:
            self.headers = DEFAULT_HEADERS

        # Define LIMS.
        self.lims = lims
        if lims is None:
            self.lims = LIMS

        # Optional Google Maps integration to retrieve GIS data.
        self.google_maps_api_key = google_maps_api_key
    
        # Assign all of the parsing routines.
        if init_all:
            for values in self.lims.values():
                script = values['coa_algorithm'].replace('.py', '')
                module = f'cannlytics.data.coas.algorithms.{script}'
                entry_point = values['coa_algorithm_entry_point']
                setattr(self, entry_point, importlib.import_module(module))

    def aggregate(
            self,
            datafiles,
            output: Optional[str] = None,
            sheet_name: Optional[str] = None,
            **kwargs,
        ) -> pd.DataFrame:
        """Aggregate multiple datafiles into a single dataset.
        Args:
            datafiles (iterable): A list of filenames, DataFrames, or 
                lists of dictionaries to aggregate into a single DataFrame.
            output (str): A path or file where data should be output (optional).
            sheet_name (str): A specific sheet to read data (optional).
            kwargs (Keywords): Keywords to pass to `save`.
        Returns:
            (list): Returns a list of data dictionaries.
        """
        # Aggregate DataFrames.
        agg = pd.DataFrame()
        for d in datafiles:
            if isinstance(d, str):
                df = pd.read_excel(d, sheet_name=sheet_name)
                agg = pd.concat([agg, df])
            elif isinstance(d, pd.DataFrame):
                agg = pd.concat([agg, d])
            elif isinstance(d, list):
                agg = pd.concat([agg, pd.DataFrame(d)])

        # Replace NaN values with `None`.
        agg = agg.where(pd.notnull(agg), None)

        # Remove duplicates.
        agg.drop_duplicates('sample_hash', inplace=True)

        # Create aggregation columns.
        agg['results'] = agg['results'].apply(dump_column)
        agg = agg.loc[agg['results'].notnull()]

        # TODO: Also handle `images` and `coa_urls`.
        # agg['images'] = agg['images'].apply(dump_column)
        # agg['coa_urls'] = agg['coa_urls'].apply(dump_column)

        # Re-hash the dataset.
        datafile_hash = create_hash(agg)

        # Save the data.
        datafile = None
        if os.path.isdir(output):
            datafile = f'{output}/{datafile_hash}.xlsx'
        elif os.path.isfile(output):
            datafile = output
        elif output is True:
            datafile = f'{datafile_hash}.xlsx'
        if datafile:
            self.save(agg, datafile, **kwargs)
        return agg

    def decode_pdf_qr_code(
            self,
            page: Any,
            img: Any,
            resolution: Optional[int] = 300,
        ) -> list:
        """Decode a PDF QR Code from a given image.
        Args:
            page (Page): A pdfplumber Page containing the image.
            img (Image): A pdfplumber Image.
            resolution (int): The resolution to render the QR code,
                `300` by default (optional).
        Returns:
            (list): The QR code data.
        """
        y = page.height
        bbox = (img['x0'], y - img['y1'], img['x1'], y - img['y0'])
        crop = page.crop(bbox)
        obj = crop.to_image(resolution=resolution)
        return pyzbar.decode(obj.original)

    def find_pdf_qr_code_url(
            self,
            pdf: Any,
            image_index: Optional[int] = None,
            page_index: Optional[int] = 0,
            resolution: Optional[int] = 300,
        ) -> str:
        """Find the QR code given a CoA PDF or page.
        If no `image_index` is provided, then all images are tried to be
        decoded until a QR code is found. If no QR code is found, then a
        `IndexError` is raised.
        Args:
            pdf (PDF or Page): A pdfplumber PDF or Page.
            image_index (int): A known image index for the QR code.
            page_index (int): The page to search, 0 by default (optional).
            resolution (int): The resolution to render the QR code,
                `300` by default (optional).
        Returns:
            (str): The QR code URL.
        """
        image_data = None
        if isinstance(pdf, str):
            pdf_file = pdfplumber.open(pdf)
            page = pdf_file.pages[page_index]
        elif isinstance(pdf, pdfplumber.pdf.PDF):
            page = pdf.pages[page_index]
        else:
            page = pdf
        if image_index:
            img = page.images[image_index]
            decoded_image = self.decode_pdf_qr_code(page, img, resolution)
            image_data = decoded_image[0].data.decode('utf-8')
        else:
            image_range = sandwich_list(page.images)
            for img in image_range:
                try:
                    decoded_image = self.decode_pdf_qr_code(page, img, resolution)
                    image_data = decoded_image[0].data.decode('utf-8')
                    if image_data:
                        break
                except:
                    continue
        return image_data

    def find_metrc_ids(
            self,
            doc: Any,
            prefixes: Optional[List[str]] = None,
            id_length: Optional[int] = 24,
        ) -> List[str]:
        """Find any Metrc IDs that may be in a given CoA PDF by searching
        for known Metrc prefixes and assuming that the ID is 24 characters.
        Arg:
            doc (str or PDF or Page): A filename, PDF, or Page.
            prefixes (list): A list of Metrc prefixes (optional).
            id_length (int): The length of a Metrc ID, 24 by default (optional).
        Returns:
            (list): Returns a list of Metrc IDs.
        """
        if isinstance(doc, str):
            pdf_file = pdfplumber.open(doc)
            pages = pdf_file.pages
        elif isinstance(doc, pdfplumber.pdf.Page):
            pages = [doc]
        else:
            pages = doc.pages
        if prefixes is None:
            prefixes = METRC_PREFIXES
        metrc_ids = []
        for prefix in prefixes:
            for page in pages:
                text = page.extract_text().replace('\n', '')
                words = text.split(' ')
                ids = [x for x in words if x.startswith(prefix) and \
                    len(x) == id_length]
                metrc_ids.extend(ids)
        return list(set(metrc_ids))

    def get_metrc_results(self, metrc_id: str) -> dict:
        """Get Metrc lab results that have been archived
        in the Cannlytics library via the Cannlytics API.
        Args:
            metrc_id (str): A Metrc UID.
        Returns:
            (dict): The sample data.
        """
        url = f'https://cannlytics.com/api/data/results?metrc_id={metrc_id}'
        response = requests.get(url)
        try:
            return response.json()['data']
        except (json.decoder.JSONDecodeError, KeyError):
            return None

    def get_page_rows(self, page: Any, **kwargs) -> list:
        """Get the rows a given page.
        Args:
            page (Page): A pdfplumber page containing rows to extract.
        Returns:
            (list): A list of text.
        """
        txt = page.extract_text(**kwargs)
        txt = txt.replace('\xa0\xa0', '\n').replace('\xa0', ',')
        return txt.split('\n')

    def get_pdf_creation_date(self, pdf: Any) -> str:
        """Get the creation date of a PDF in ISO format.
        Args:
            pdf (PDF): A pdfplumber PDF.
        Returns:
            (str): An ISO formatted date.
        """
        if isinstance(pdf, str):
            pdf_file = pdfplumber.open(pdf)
        else:
            pdf_file = pdf
        try:
            date = pdf_file.metadata['CreationDate'].split('D:')[-1]
        except KeyError:
            return None
        isoformat = f'{date[0:4]}-{date[4:6]}-{date[6:8]}'
        isoformat += f'T{date[8:10]}:{date[10:12]}:{date[12:14]}'
        return isoformat

    def get_pdf_image_data(
            self,
            page: Any,
            bbox: Optional[tuple] = None,
            image_index: Optional[int] = 0, 
            resolution: Optional[int] = 300,
        ) -> str:
        """Get the image data for a given PDF page image.
        Args:
            page (Page): A pdfplumber Page.
            image_index (int): The index of the image, 0 by default (optional).
            resolution (int): The resolution for the image, 300 by default (optional).
        Returns:
            (str): The image data.
        """
        if bbox is None:
            y = page.height
            img = page.images[image_index]
            bbox = (img['x0'], y - img['y1'], img['x1'], y - img['y0'])
        crop = page.crop(bbox)
        obj = crop.to_image(resolution=resolution)
        buffered = BytesIO()
        obj.save(buffered, format='JPEG')
        img_str = base64.b64encode(buffered.getvalue())
        return img_str.decode('utf-8')

    def save_image_data(
            self,
            image_data: str,
            image_file: Optional[str] ='image.png',
        ):
        """Save image data to a file.
        Args:
            image_data (str): The image data.
            image_file (str): A filename for the image.
        """
        image_bytes = base64.b64decode(image_data)
        image_io = BytesIO(image_bytes)
        image = Image.open(image_io)
        image.save(image_file)

    def identify_lims(
            self,
            doc: Any,
            lims: Optional[Any] = None,
            temp_path: Optional[str] = '/tmp',
        ) -> str:
        """Identify if a CoA was created by a common LIMS.
        Search all of the text of the LIMS name or URL.
        If no LIMS is identified from the text, then the images
        are attempted to be decoded, searching for a QR code URL.
        Args:
            doc (str, PDF or Page): A URL or a pdfplumber PDF or Page.
            lims (str or dict): The name of a specific LIMS or a
                dictionary of known LIMS.
            temp_path (str): A temporary directory to store any online PDFs
                if needed for identification, `/tmp` by default (optional).
        Returns:
            (str): Returns LIMS name if found, otherwise returns `None`.
        """
        # Search all of the text of the LIMS name or URL.
        known = None
        if isinstance(doc, str):

            # Handle shortened URLs.
            text = doc
            if doc.startswith('https://tinyurl'):
                response = requests.get(doc, headers=DEFAULT_HEADERS)
                text = response.url
        
            # Handle Google Drive PDFs.
            if text.startswith('https://drive.google'):
                temp_pdf = os.path.join(temp_path, 'coa.pdf')
                if not os.path.exists(temp_path): os.makedirs(temp_path)
                download_google_drive_file(text, temp_pdf)
                text = temp_pdf

            # Handle PDFs.
            try:
                pdf_file = pdfplumber.open(text)
                text = pdf_file.pages[0].extract_text()
            except (FileNotFoundError, OSError):
                pass
        
        # Handle PDFs passed directly.
        else:
            if isinstance(doc, pdfplumber.pdf.PDF):
                page = doc.pages[0]
            else:
                page = doc
            text = page.extract_text()

        # Handle custom LIMS.
        if lims is None:
            lims = LIMS
        
        # Iterate over all known LIMS looking for URLs and LIMS names.
        if isinstance(lims, str):
            try:
                if self.lims[lims]['url'] in text:
                    known = lims
            except KeyError:
                if lims.lower() in text.lower():
                    known = lims
        else:
            for key, values in lims.items():
                url = values.get('url')
                if url and url in text:
                    known = key
                    break
                lab = values.get('lims')
                if lab and lab in text:
                    known = key
                    break

        # If the LIMS is still unknown, then try to identify the LIMS
        # with any a QR code, if the COA is a PDF.
        if not known:
            try:
                qr_code_url = self.find_pdf_qr_code_url(doc)
            except:
                qr_code_url = None
            if qr_code_url:
                for key, values in lims.items():
                    url = values.get('url')
                    if url and url in qr_code_url:
                        known = key
                        break
        
        # Optional: Can we detect LIMS from a photo for a COA or a label?

        # Return any known LIMS.
        return known

    def parse(
            self,
            data: Any,
            cleanup: Optional[bool] = True,
            headers: Optional[dict] = {},
            kind: Optional[str] = 'url',
            lims: Optional[Any] = None,
            max_delay: Optional[float] = 7,
            persist: Optional[bool] = True,
            resolution: Optional[int] = 300,
            temp_path: Optional[str] = '/tmp',
            use_cached: Optional[bool] = False,
        ) -> list:
        """Parse all CoAs given a directory, a list of files,
        or a list of URLs.
        Args:
            data (str or list): A directory (str) or a list
                of PDF file paths or a list of CoA URLs.
            cleanup (bool): Whether or not to remove the files generated
                during OCR, `True` by default (optional).
            headers (dict): Headers for HTTP requests (optional).
            kind (str): The kind of CoA input, PDF or URL, `url`
                by default (optional).
            lims (str or dict): Specific LIMS to parse CoAs (optional).
            max_delay (float): The maximum number of seconds to wait
                for the page to load (optional).
            persist (bool): Whether to persist the driver
                and / or session between CoA parses, the
                default is `True`, with any driver and session
                being closed at the end (optional).
            temp_path (str): A temporary directory to store files used
                during PDF OCR, `/tmp` by default (optional).
            resolution (int): The resolution of rendered PDF images,
                300 by default (optional).
        Returns:
            (list): Returns a list of all of the PDFs.
        """
        coas, docs = [], []

        # Parse a URL, PDF path, or .zip folder.
        if isinstance(data, str):

            # FIXME: Parse a URL to an image of a COA or QR code.
            image_extensions = ['.png', '.jpg', '.jpeg']
            if 'https' in data and any(ext in data.lower() for ext in image_extensions):
                coa_data = self.parse_url(
                    data,
                    headers=headers,
                    lims=lims,
                    max_delay=max_delay,
                    persist=persist,
                )
                coas.append(coa_data)

            # Parse a URL.
            if 'https' in data:
                coa_data = self.parse_url(
                    data,
                    headers=headers,
                    lims=lims,
                    max_delay=max_delay,
                    persist=persist,
                )
                coas.append(coa_data)

            # Parse a PDF.
            elif '.pdf' in data:
                coa_data = self.parse_pdf(
                    data,
                    cleanup=cleanup,
                    headers=headers,
                    lims=lims,
                    max_delay=max_delay,
                    persist=persist,
                    resolution=resolution,
                    temp_path=temp_path,
                    use_cached=use_cached,
                )
                coas.append(coa_data)

            # TODO: Handle paths to images of COAs and QR codes.

            # Parse a ZIP.
            elif '.zip' in data:
                doc_dir = unzip_files(data)
                docs = get_directory_files(doc_dir)

            # Parse a directory.
            else:
                docs = get_directory_files(data)

        # Handle a list of URLs, PDFs, and/or ZIPs.
        else:
            docs = data

        # Parse all of the PDFs.
        for doc in docs:

            # Parse a .zip folder.
            if '.zip' in doc and kind != 'url':
                doc_dir = unzip_files(doc)
                pdf_files = get_directory_files(doc_dir)
                for pdf_file in pdf_files:
                    coa_data = self.parse_pdf(
                        pdf_file,
                        cleanup=cleanup,
                        headers=headers,
                        lims=lims,
                        max_delay=max_delay,
                        persist=persist,
                        resolution=resolution,
                        temp_path=temp_path,
                        use_cached=use_cached,
                    )
                    coas.append(coa_data)

            # Parse a PDF.
            elif '.pdf' in doc and kind != 'url':
                coa_data = self.parse_pdf(
                    doc,
                    cleanup=cleanup,
                    headers=headers,
                    lims=lims,
                    max_delay=max_delay,
                    persist=persist,
                    resolution=resolution,
                    temp_path=temp_path,
                    use_cached=use_cached,
                )

            # TODO: Parse image of COAs and QR codes.

            # Parse a URL.
            else:
                coa_data = self.parse_url(
                    doc,
                    headers=headers,
                    lims=lims,
                    max_delay=max_delay,
                    persist=persist,
                )
            
            # Record the COA data.
            coas.append(coa_data)

        # Close the parser at the end.
        if persist:
            self.quit()

        # Return the parsed COAs.
        return coas

    def parse_pdf(
            self,
            pdf: Any,
            cleanup: Optional[bool] = True,
            headers: Optional[dict] = {},
            lims: Optional[Any] = None,
            max_delay: Optional[float] = 7,
            persist: Optional[bool] = False,
            resolution: Optional[int] = 300,
            temp_path: Optional[str] = '/tmp',
            use_cached: Optional[bool] = False,
        ) -> dict:
        """Parse a CoA PDF. Searches the best guess image, then all
        images, for a QR code URL to find results online.
        Args:
            pdf (PDF): A file path to a PDF or a pdfplumber PDF.
            cleanup (bool): Whether or not to remove the files generated
                during OCR, `True` by default (optional).
            headers (dict): Headers for HTTP requests (optional).
            lims (str or dict): Specific LIMS to parse CoAs (optional).
            max_delay (float): The maximum number of seconds to wait
                for the page to load (optional).
            persist (bool): Whether to persist the driver
                and / or session between CoA parses, the
                default is `True`, with any driver and session
                being closed at the end (optional).
            temp_path (str): A temporary directory used for OCR.
            temp_path (str): A temporary directory to store files used
                during PDF OCR, `/tmp` by default (optional).
            resolution (int): The resolution of rendered PDF images,
                300 by default (optional).
        Returns:
            (dict): The sample data.
        """
        if lims is None:
            lims = self.lims

        # Read the PDF.
        if isinstance(pdf, str):
            pdf_file = pdfplumber.open(pdf)
        elif isinstance(pdf, pdfplumber.pdf.PDF):
            pdf_file = pdf
        else:
            with open(magick_wand(file=pdf, resolution=resolution)) as temp_file:
                temp_file.save(f'{temp_path}/coa.pdf')
            pdf_file = pdfplumber.open(f'{temp_path}/coa.pdf')

        # Optional: Try to find any Metrc IDs to query the Cannlytics API.
        # Return cached results instead of parsing the CoA from scratch.
        if use_cached is True:
            metrc_ids = self.find_metrc_ids(pdf_file.pages[0])
            if metrc_ids:
                for metrc_id in metrc_ids:
                    data = self.get_metrc_results(metrc_id)
                    if data is not None:
                        return data

        # Identify any known LIMS, trying OCR if the PDF is not recognized,
        # then raise an error if the labs / LIMS is unknown for safety.
        # TODO: Parse unidentified CoAs to the best of our abilities.
        known_lims = self.identify_lims(pdf_file, lims=lims)
        if known_lims is None:
            temp_pdf = f'{temp_path}/ocr-coa.pdf'
            if isinstance(pdf, str):
                filename = pdf
            elif isinstance(pdf, pdfplumber.pdf.PDF):
                filename = pdf.stream.name
            else:
                filename = f'{temp_path}/coa.pdf'
            self.pdf_ocr(
                filename,
                temp_pdf,
                temp_path=temp_path,
                resolution=resolution,
                cleanup=cleanup,
            )
            pdf_file = pdfplumber.open(temp_pdf)
            known_lims = self.identify_lims(pdf_file, lims=lims)
            if known_lims is None:
                raise ValueError(UNIDENTIFIED_LIMS)

        # Get the time the CoA was created, if known.
        date_tested = self.get_pdf_creation_date(pdf_file)

        # Attempt to use an URL from any QR code on the PDF.
        url = None
        try:
            qr_code_index = self.lims[known_lims].get('qr_code_index')
            url = self.find_pdf_qr_code_url(pdf_file, qr_code_index)
            if url is None and qr_code_index is not None:
                url = self.find_pdf_qr_code_url(pdf_file)
        except IndexError:
            url = self.find_pdf_qr_code_url(pdf_file)

        # Get the LIMS parsing routine.
        algorithm_name = LIMS[known_lims]['coa_algorithm_entry_point']
        algorithm = getattr(getattr(self, algorithm_name), algorithm_name)

        # Use the URL if found, then try the PDF if the URL fails or is missing.
        if url:
            try:
                data = self.parse_url(
                    url,
                    headers=headers,
                    lims=known_lims,
                    max_delay=max_delay,
                    persist=persist,
                )
            except:
                data = algorithm(
                    self,
                    pdf_file,
                    headers=headers,
                    max_delay=max_delay,
                    persist=persist,
                    google_maps_api_key=self.google_maps_api_key,
                )
        else:
            data = algorithm(
                self,
                pdf_file,
                headers=headers,
                max_delay=max_delay,
                persist=persist,
                google_maps_api_key=self.google_maps_api_key,
            )

        pdf_file.close()
        sample = {
            'date_tested': date_tested,
            'lab_results_url': url,
            'lims': known_lims,
        }
        return {**sample, **data}

    def parse_url(
            self,
            url: str,
            lims: Optional[Any] = None,
            headers: Optional[dict] = {},
            max_delay: Optional[float] = 7,
            persist: Optional[bool] = False,
        ) -> dict:
        """Parse a CoA URL.
        Args:
            url (str): The CoA URL.
            lims (str or dict): Specific LIMS to parse CoAs.
            headers (Any): Optional headers for standardization.
            max_delay (float): The maximum number of seconds to wait
                for the page to load.
            persist (bool): Whether to persist the session.
                The default is `False`. If you do persist
                the driver, then make sure to call `quit`
                when you are finished.
        Returns:
            (dict): The sample data.
        """
        if lims is None:
            lims = self.lims

        # Identify the LIMS.
        known_lims = self.identify_lims(url, lims=lims)

        # Restrict to known labs / LIMS for safety.
        # TODO: Parse unidentified CoAs to the best of our abilities.
        if known_lims is None:
            raise ValueError(UNIDENTIFIED_LIMS)

        # Get the LIMS parsing routine.
        algorithm_name = LIMS[known_lims]['coa_algorithm_entry_point']
        algorithm = getattr(getattr(self, algorithm_name), algorithm_name)
        data = algorithm(
            self,
            url,
            headers=headers,
            max_delay=max_delay,
            persist=persist,
            google_maps_api_key=self.google_maps_api_key,
        )
        return data

    def pdf_ocr(
            self,
            filename: str,
            outfile: str,
            temp_path: Optional[str] = '/tmp',
            resolution: Optional[int] = 300,
            cleanup: Optional[bool] = True,
        ) -> None:
        """Pass a PDF through OCR to recognize its text. Outputs a new PDF.
        A temporary directory is used, because the algorithm is to:
            1. Convert all PDF pages to images.
            2. Convert each image to PDF with text.
            3. Compile the PDFs with text to a single PDF.
        The rendered images and individual PDF files are removed by default.
        Args:
            filename (str): The filename of the PDF to apply OCR.
            outfile (str): A new PDF file to generate.
            temp_path (str): A temporary directory to store files used
                during PDF OCR, `/tmp` by default (optional).
            resolution (int): The resolution of rendered PDF images,
                300 by default (optional).
            cleanup (bool): Whether or not to remove the files generated
                during OCR, `True` by default (optional).
        """
        # Create a directory to store images and rendered PDFs.
        if not os.path.exists(temp_path): os.makedirs(temp_path)

        # Convert each PDF page to an image.
        image_files = convert_pdf_to_images(
            filename,
            temp_path,
            resolution=resolution,
        )

        # Convert each image to PDF with text.
        pdf_files = []
        for image_file in image_files:
            pdf = image_to_pdf_or_hocr(image_file, extension='pdf')
            pdf_file = image_file.replace('.png', '.pdf')
            pdf_files.append(pdf_file)
            with open(pdf_file, 'w+b') as f:
                f.write(pdf)
                f.close()
            if cleanup:
                os.remove(image_file)

        # Compile the PDFs with text to a single PDF.
        merger = PdfMerger()
        for pdf in pdf_files:
            merger.append(pdf)
        merger.write(outfile)
        merger.close()

        # Remove individual PDF files.
        if cleanup:
            for pdf in pdf_files:
                os.remove(pdf)

        # Remove all `magick-*` files from the temp directory.
        # for i in os.listdir(temp_path):
        #     path = os.path.join(temp_path, i)
        #     if os.path.isfile(path) and i.startswith('magick-'):
        #         os.remove(path)

    def save(
            self,
            data: Any,
            outfile:str,
            codings: Optional[dict] = None,
            column_order: Optional[list] = None,
            nuisance_columns: Optional[list] = None,
            numeric_columns: Optional[list] = None,
            standard_analyses: Optional[dict] = None,
            standard_analytes: Optional[dict] = None,
            standard_fields: Optional[dict] = None,
            google_maps_api_key: Optional[str] = None,
        ) -> Any:
        """Save all CoA data, elongating results and widening values.
        That is, a Workbook is created with a "Details" worksheet that
        has all of the raw data, a "Results" worksheet with long-form
        data where each row is a result for an analyte, and a "Values"
        worksheet with wide-form data where each row is an observation
        and each column is the `value` field for each of the `results`.
        Args:
            data (dict or list or DataFrame): The data to save.
            outfile (str): The file that you wish to save. Accepts a
                response object for returning in an HTTP request.
            codings (dict): A map of value codings, from actual to coding.
            column_order (list): Desired order for columns.
            nuisance_columns (list): A list of column suffixes to remove.
            numeric_columns (list): A list of columns to treat as numeric
                and apply codings.
            standard_analyses (dict): A mapping of encountered analyses to
                standard analyses.
            standard_analytes (dict): A mapping of encountered analytes to
                standard analytes.
            standard_fields (dict): A mapping of encountered fields to
                standard fields.
            google_maps_api_key (str): A Google Maps API Key to supplement
                addresses with latitude and longitude.
        Returns:
            (Workbook): An openpyxl Workbook.
        """
        # Create a workbook for saving the data.
        wb = openpyxl.Workbook()

        # Initialize the details data.
        details_data = None
        if isinstance(data, dict):
            details_data = pd.DataFrame([data])
        elif isinstance(data, list):
            details_data = pd.DataFrame(data)
        else:
            details_data = data

        # Specify the desired order for columns / fields.
        if codings is None:
            codings = self.codings

        # Standardize details.
        details_data = self.standardize(
            details_data,
            column_order=column_order,
            nuisance_columns=nuisance_columns,
            numeric_columns=numeric_columns,
            standard_analyses=standard_analyses,
            standard_analytes=standard_analytes,
            standard_fields=standard_fields,
            google_maps_api_key=google_maps_api_key,
        )

        # Standardize results.
        results_data = self.standardize(
            details_data,
            how='long',
            column_order=column_order,
            nuisance_columns=nuisance_columns,
            numeric_columns=numeric_columns,
            standard_analyses=standard_analyses,
            standard_analytes=standard_analytes,
            standard_fields=standard_fields,
            google_maps_api_key=google_maps_api_key,
        )

        # Standardize values.
        values_data = self.standardize(
            details_data,
            how='wide',
            details_data=details_data,
            results_data=results_data,
            column_order=column_order,
            nuisance_columns=nuisance_columns,
            numeric_columns=numeric_columns,
            standard_analyses=standard_analyses,
            standard_analytes=standard_analytes,
            standard_fields=standard_fields,
            google_maps_api_key=google_maps_api_key,
        )

        # Add a "Codings" worksheet.
        coding_data = pd.DataFrame({
            'Coding': codings.values(),
            'Actual': codings.keys(),
        })
        
        # Format details `results`, `coa_urls`, `images` as proper JSON.
        try:
            details_data['results'] = details_data['results'].apply(json.dumps)
        except:
            pass
        try:
            details_data['coa_urls'] = details_data['coa_urls'].apply(json.dumps)
        except:
            pass
        try:
            details_data['images'] = details_data.get['images'].apply(json.dumps)
        except:
            pass

        # Create a workbook for saving the data.
        wb = openpyxl.Workbook()
        ws = wb.worksheets[0]
        ws.title = 'Values'
        write_to_worksheet(ws, values_data)
        wb.create_sheet(title='Details')
        ws = wb.worksheets[1]
        write_to_worksheet(ws, details_data)
        wb.create_sheet(title='Results')
        ws = wb.worksheets[2]
        write_to_worksheet(ws, results_data)
        wb.create_sheet(title='Codings')
        ws = wb.worksheets[3]
        write_to_worksheet(ws, coding_data)
        wb.save(outfile)
        wb.close()
        return wb

    def standardize(
            self,
            data: Any,
            codings: Optional[dict] = None,
            column_order: Optional[list] = None,
            nuisance_columns: Optional[list] = None,
            numeric_columns: Optional[list] = None,
            how: Optional[str] = 'details',
            details_data: Optional[Any] = None,
            results_data: Optional[Any] = None,
            standard_analyses: Optional[dict] = None,
            standard_analytes: Optional[dict] = None,
            standard_fields: Optional[dict] = None,
            google_maps_api_key: Optional[str] = None,
        ) -> Any:
        """Standardize (and normalize) given data.
        Args:
            data (dict or list or DataFrame): The data to standardize.
            codings (dict): A map of value codings, from actual to coding.
            column_order (list): A list of columns in desired order.
            nuisance_columns (list): A list of column suffixes to remove.
            numeric_columns (list): A list of columns to treat as numeric
                and apply codings.
            standard_analyses (dict): A mapping of encountered analyses to
                standard analyses.
            standard_analytes (dict): A mapping of encountered analytes to
                standard analytes.
            standard_fields (dict): A mapping of encountered fields to
                standard fields.
            how (str): How to standardize, a simple clean of the data
                `details` by default. Alternatively specify `wide` for a
                wide-form DataFrame of values or `long` for a long-form
                DataFrame of results.
            details_data (DataFrame): The data pre-formatted as `details`.
                May provide a speed increase provided when formatting `values`.
            results_data (DataFrame): The data pre-formatted as `results`.
                May provide a speed increase provided when formatting `values`.
            google_maps_api_key (str): A Google Maps API Key to supplement
                addresses with latitude and longitude.
        Returns:
            (dict or list or DataFrame): Returns the data with standardized
                fields, analyses, analytes, product types, normalized
                results, and augmented with strain name and GIS data.
        """
        # Specify the desired order for columns / fields.
        if codings is None:
            codings = self.codings
        if column_order is None:
            column_order = self.column_order
        if nuisance_columns is None:
            nuisance_columns = self.nuisance_columns
        if numeric_columns is None:
            numeric_columns = self.numeric_columns
        if standard_analyses is None:
            standard_analyses = self.analyses
        if standard_analytes is None:
            standard_analytes = self.analytes
        if standard_fields is None:
            standard_fields = self.fields

        # Future standardization work:
        # [ ] Calculate all missing totals:`total_cannabinoids`, `total_terpenes`, etc.
        # [ ] Remove and keep `units` from `value`.
        # [ ] Standardize `units`
        # [ ] Create a standard `product_type_key`
        # [ ] Augment any missing GIS data, such as latitude, longitude, or address field.
        # [ ] Try to parse a `strain_name` from `product_name` with NLP.

        # Standardize a dictionary.
        if isinstance(data, dict):

            # Identify standard fields, adding analytes for `wide` data.
            fields = standard_fields
            if how == 'wide':
                current_fields = {x:x for x in data.keys()}
                fields = {**current_fields, **standard_analytes}

            # Standardize fields.
            std = {}
            for k, v in data.items():
                key = fields.get(k, k)
                std[key] = v

            # Standardize `analyses` of details.
            if how == 'details':
                std['analyses'] = [standard_analyses.get(x, x) for x in std['analyses']]

                # Standardize the `analysis` of reach of the `results`.
                # Also, normalize the `results`, converting to numeric
                # and applying codings.
                standardized_results = []
                sample_results = data['results']
                if isinstance(sample_results, str):
                    try:
                        sample_results = ast.literal_eval(sample_results)
                        if isinstance(sample_results, str):
                            sample_results = json.load(json.dumps(sample_results))
                    except:
                        json.loads(sample_results)
                for result in sample_results:
                    
                    analysis = result.get('analysis')
                    result['analysis'] = standard_analyses.get(analysis, analysis)
                    for c in numeric_columns:
                        value = result.get(c)
                        value = codings.get(value, value)
                        if isinstance(value, str):
                            value = convert_to_numeric(value, strip=True)
                        result[c] = pd.to_numeric(value, errors='coerce')
                    standardized_results.append(result)
                std['results'] = standardized_results
                
            # Standard `analysis` of long-form data.
            elif how == 'long':
                analysis = std.get('analysis')
                std['analysis'] = standard_analyses.get(analysis, analysis)

            # Normalize the numeric fields for `wide` data.
            if how == 'wide':
                analytes = list([a for a in data.keys() if a not in column_order])
                for c in analytes:
                    value = std.get(c)
                    try:
                        value = codings.get(value, value)
                    except TypeError:
                        pass
                    if isinstance(value, str):
                        value = convert_to_numeric(value, strip=True)
                    std[c] = pd.to_numeric(value, errors='coerce')

            # Normalize numeric fields for `wide` data.
            elif how == 'long':
                for c in numeric_columns:
                    value = std.get(c)
                    value = codings.get(value, value)
                    if isinstance(value, str):
                        value = convert_to_numeric(value, strip=True)
                    std[c] = pd.to_numeric(value, errors='coerce')

            # Turn dates values to ISO format.
            dates = [x for x in std.keys() if x.startswith('date')]
            for k in dates:
                try:
                    std[k] = pd.to_datetime(std[k]).isoformat()
                except:
                    pass

            # Return the standardized observation.
            return std

        # Standardize a list of dictionaries, series, or DataFrames.
        elif isinstance(data, list):
            return [self.standardize(
                x,
                how=how,
                codings=codings,
                column_order=column_order,
                nuisance_columns=nuisance_columns,
                numeric_columns=numeric_columns,
                standard_analyses=standard_analyses,
                standard_analytes=standard_analytes,
                standard_fields=standard_fields,
                google_maps_api_key=google_maps_api_key,
            ) for x in data]

        # Standardize a DataFrame.
        elif isinstance(data, pd.DataFrame):

            # Standardize details (`details` data).
            if how == 'details':

                # Standardize detail columns.
                details_data = data.copy(deep=True)
                details_data.rename(
                    standard_fields,
                    axis=1,
                    inplace=True,
                    errors='ignore',
                )
                details_data = details_data.groupby(level=0, axis=1).first()

                # Drop nuisance columns.
                for c in nuisance_columns:
                    criterion = details_data.columns.str.endswith(c)
                    details_data = details_data.loc[:, ~criterion]

                # Apply codings to results.
                # FIXME: This is super slow (2-3 mins for 2.5k observations)!
                if 'results' in details_data.columns:
                    details_data['results'].replace(codings, inplace=True)
                # except KeyError:
                #     pass

                # Convert totals to numeric.
                # TODO: Calculate totals if they don't already exist:
                totals = [x for x in details_data.keys() if x.startswith('total_')]
                for c in totals:
                    details_data[c] = details_data[c].astype(str).apply(convert_to_numeric, strip=True)
                    details_data[c] = details_data[c].apply(pd.to_numeric, errors='coerce')

                # Re-order columns.
                details_data = reorder_columns(details_data, column_order)
                return details_data            

            # Standardize results (`long` data).
            elif how == 'long':

                # Create a long table of results data.
                results = []
                details_data = data.copy(deep=True)
                for _, item in details_data.iterrows():

                    # Get the sample results.
                    sample_results = item['results']
                    # if isinstance(sample_results, str):
                    try:
                        sample_results = ast.literal_eval(sample_results)
                        if isinstance(sample_results, str):
                            sample_results = json.load(json.dumps(sample_results))
                    except:
                        try:
                            sample_results = json.loads(sample_results)
                        except:
                            pass

                    # Add each entry.
                    for result in sample_results:
                        std = {}
                        for c in column_order:
                            std[c] = item.get(c)
                        try:
                            results.append({**std, **result})
                        except TypeError:
                            results.append({**std, **json.loads(result)})

                # Apply codings (redundant?).
                results_data = pd.DataFrame(results)
                for c in numeric_columns:
                    try:
                        results_data[c] = results_data[c].replace(codings)
                    except KeyError:
                        pass

                # Standardize the results columns.
                results_data.rename(
                    standard_fields,
                    axis=1,
                    inplace=True,
                    errors='ignore',
                )
                results_data = results_data.groupby(level=0, axis=1).first()

                # Standardize the key column (redundant?).
                try:
                    results_data['key'] = results_data['key'].apply(
                        lambda x: ANALYTES.get(x, x)
                    )
                except KeyError:
                    try:
                        results_data['key'] = results_data['name'].apply(
                            lambda x: ANALYTES.get(snake_case(x), snake_case(x))
                        )
                    except KeyError:
                        pass 

                # Re-order columns.
                results_data = reorder_columns(results_data, column_order)
                return results_data

            # Standardize values (`wide` data).
            elif how == 'wide':

                # Get details and results data (if not passed for speed).
                if details_data is None:
                    details_data = self.standardize(
                        data,
                        codings=codings,
                        column_order=column_order,
                        nuisance_columns=nuisance_columns,
                        numeric_columns=numeric_columns,
                        standard_analyses=standard_analyses,
                        standard_analytes=standard_analytes,
                        standard_fields=standard_fields,
                        google_maps_api_key=google_maps_api_key,
                    )
                if results_data is None:
                    results_data = self.standardize(
                        data,
                        how='wide',
                        codings=codings,
                        column_order=column_order,
                        nuisance_columns=nuisance_columns,
                        numeric_columns=numeric_columns,
                        standard_analyses=standard_analyses,
                        standard_analytes=standard_analytes,
                        standard_fields=standard_fields,
                        google_maps_api_key=google_maps_api_key,
                    )

                # Map keys to analysis for ordering for Values worksheet columns.
                # FIXME: Handle observations without results.
                pairs = []
                analytes = list(results_data['key'].unique())
                for a in analytes:
                    try:
                        analyses = results_data.loc[results_data['key'] == a]
                        analyses = list(analyses['analysis'].unique())
                        analyses = [x for x in analyses if x is not None]
                        analysis = analyses[0]
                        place = ord(analysis[0])
                        # Hot-fix to order terpenes right after cannabinoids.
                        if place == 116:
                            place = 100 
                        pairs.append((a, analysis, place))
                    except:
                        # Hot-fix to place unidentified analyses at the end.
                        pairs.append((a, None, 122))

                # Sort the pairs of analyses / analytes.
                pairs.sort(key=operator.itemgetter(2))

                # Create a wide table of values data.
                values = []
                for _, item in details_data.iterrows():

                    # Use all of the details in the values.
                    std = item.to_dict()

                    # Old: Only use the default columns. Make optional?
                    # std = {}
                    # for c in column_order:
                    #     std[c] = item[c]

                    # Get the sample results.
                    sample_results = item['results']
                    if isinstance(sample_results, str):
                        try:
                            sample_results = ast.literal_eval(sample_results)
                            if isinstance(sample_results, str):
                                sample_results = json.load(json.dumps(sample_results))
                        except:
                            sample_results = json.loads(sample_results)

                    # Keep the values from each result.
                    for result in sample_results:
                        result = {k: v for k, v in result.items() if v == v}
                        analyte = result.get('key', snake_case(result.get('name')))
                        analyte = standard_analytes.get(analyte, analyte)
                        value = result.get('value', result.get('percent', result.get('mg_g')))
                        std[analyte] = value
                    values.append(std)

                # Rename and combine columns.
                values_data = pd.DataFrame(values)
                values_data.rename(
                    standard_analytes,
                    axis=1,
                    inplace=True,
                    errors='ignore',
                )
                values_data = values_data.groupby(level=0, axis=1).first()

                # Apply codings to columns.
                values_data.replace(codings, inplace=True)

                # Drop nuisance columns.
                for c in nuisance_columns:
                    criterion = values_data.columns.str.endswith(c)
                    values_data = values_data.loc[:, ~criterion]

                # Old: Drop totals. Make optional?
                # criterion = values_data.columns.str.startswith('total_')
                # values_data = values_data.loc[:, ~criterion]

                # Move certain columns to the beginning.
                cols = list(details_data.columns) + [x[0] for x in pairs]
                values_data = reorder_columns(values_data, cols)
                return values_data

        # Standardize a series.
        elif isinstance(data, pd.Series):
            return pd.Series(self.standardize(
                data.to_dict(),
                how=how,
                codings=codings,
                column_order=column_order,
                nuisance_columns=nuisance_columns,
                numeric_columns=numeric_columns,
                standard_analyses=standard_analyses,
                standard_analytes=standard_analytes,
                standard_fields=standard_fields,
                google_maps_api_key=google_maps_api_key,
            ))

        # Raise an error if an incorrect type is passed.
        else:
            raise ValueError
    
    def parse_with_ai(
            self,
            filename: Any,
            temp_path: Optional[str] = None,
            session: Optional[Any] = None,
            headers: Optional[dict] = DEFAULT_HEADERS,
            use_cached: Optional[bool] = False,
            openai_api_key: Optional[str] = None,
            model: Optional[str] = 'gpt-4',
            max_tokens: Optional[int] = 4_000,
            temperature: Optional[float] = 0.0,
            initial_cost: Optional[float] = 0.0,
            instructional_prompt: Optional[str] = None,
            results_prompt: Optional[str] = None,
            coa_prompt: Optional[str] = None,
            max_prompt_length: Optional[int] = 4_000,
            verbose: Optional[bool] = False,
            user: Optional[str] = None,
            retry_pause: Optional[float] = 3.33,
        ) -> list:
        """Parse a COA with AI.
        Args:
            filename (str): The filename or URL of the COA to parse.
        Returns:
            (dict): The parsed CoA data.
        """
        data, prompts, cost = parse_coa_with_ai(
            self,
            filename,
            temp_path=temp_path,
            session=session,
            headers=headers,
            use_cached=use_cached,
            openai_api_key=openai_api_key,
            model=model,
            max_tokens=max_tokens,
            temperature=temperature,
            initial_cost=initial_cost,
            instructional_prompt=instructional_prompt,
            results_prompt=results_prompt,
            coa_prompt=coa_prompt,
            max_prompt_length=max_prompt_length,
            verbose=verbose,
            user=user,
            retry_pause=retry_pause,
        )
        return data, prompts, cost
    
    def scan(
            self,
            filename: Any,
            width: Optional[int] = 1024,
            temp_path: Optional[str] = '/tmp'
        ) -> str:
        """Scan an image for a QR code or barcode and return any data.
        Args:
            filename (str): A path to an image with a barcode or qr code.
            width (int): The base width to resize the image.
        Returns:
            (str): Returns the data from the decoded QR code.
        """
        # Handle the filename.
        if isinstance(filename, str):
            image = cv2.imread(filename)
        elif isinstance(filename, np.ndarray):
            image = filename
        else:
            try:
                image = cv2.imread(filename.filename)
            except:
                raise ValueError('`filename` must be a string or Image.')

        # If the temp path has any extension, then use it as the outfile.
        if os.path.splitext(temp_path)[1] != '':
            outfile = temp_path

        # Otherwise, create a temporary file to store the image.
        else:
            if not os.path.exists(temp_path): os.makedirs(temp_path)
            timestamp = datetime.now().strftime("%Y%m%d%H%M%S%f")
            outfile = os.path.join(temp_path, f'{timestamp}.png')

        # Load the image, apply grayscale, Gaussian blur, and Otsu's threshold.
        # Use morphology to find and connect text contours.
        # Finally, filter for any QR code.
        qr_code_found = False
        original = image.copy()
        gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        blur = cv2.GaussianBlur(gray, (9,9), 0)
        thresh = cv2.threshold(blur, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)[1]
        kernel = cv2.getStructuringElement(cv2.MORPH_RECT, (5,5))
        close = cv2.morphologyEx(thresh, cv2.MORPH_CLOSE, kernel, iterations=2)
        contours = cv2.findContours(close, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        contours = contours[0] if len(contours) == 2 else contours[1]
        for c in contours:
            peri = cv2.arcLength(c, True)
            approx = cv2.approxPolyDP(c, 0.04 * peri, True)
            x, y, w, h = cv2.boundingRect(approx)
            area = cv2.contourArea(c)
            ar = w / float(h)
            if len(approx) == 4 and area > 1000 and (ar > .85 and ar < 1.3):
                cv2.rectangle(image, (x, y), (x + w, y + h), (36, 255, 12), 3)
                cv2.imwrite(outfile, original[y: y + h, x: x + w])
                qr_code_found = True

        # If a QR code was not found, then return None.
        if not qr_code_found:
            return None

        # If a width is given, then resize the image to facilitate
        # QR code reading. Calculates the height based on the new width
        # and the original aspect ratio.
        if width:
            img = Image.open(outfile)
            w_percent = (width / float(img.size[0]))
            h_size = int((float(img.size[1]) * float(w_percent)))
            img = img.resize((width, h_size), Image.ANTIALIAS)
            img.save(outfile)

        # Read the resized image again (important) and try to decode QR codes.
        code = None
        image = Image.open(outfile)
        codes = pyzbar.decode(image)
        if codes:
            code = codes[0].data.decode('utf-8')
        return code

    def quit(self):
        """Close any driver, end any session, and reset the parameters."""
        try:
            self.driver.quit()
        except:
            pass
        try:
            self.session.close()
        except:
            pass
        self.driver = None
        self.options = None
        self.session = None
        self.service = None


# === Tests ===
if __name__ == '__main__' and False:

    # Initialize the CoA parser.
    # Future work: Test the parser with different configurations.
    parser = CoADoc()

    # Specify where your data lives for testing.
    DATA_DIR = '../../../.datasets/coas'
    cc_coa_pdf = f'{DATA_DIR}/Classic Jack.pdf'
    cc_coa_url = 'https://share.confidentcannabis.com/samples/public/share/4ee67b54-be74-44e4-bb94-4f44d8294062'
    tagleaf_coa_pdf = f'{DATA_DIR}/Sunbeam.pdf'
    tagleaf_coa_url = 'https://lims.tagleaf.com/coas/F6LHqs9rk9vsvuILcNuH6je4VWCiFzdhgWlV7kAEanIP24qlHS'
    tagleaf_coa_short_url = 'https://lims.tagleaf.com/coa_/F6LHqs9rk9'

    # [✓] TEST: `decode_pdf_qr_code` via `find_pdf_qr_code_url`.
    qr_code_url = parser.find_pdf_qr_code_url(cc_coa_pdf)
    assert qr_code_url.startswith('https')
    qr_code_url = parser.find_pdf_qr_code_url(tagleaf_coa_pdf)
    assert qr_code_url.startswith('https')

    # [✓] TEST: `get_pdf_creation_date`.
    from datetime import datetime
    creation_date = parser.get_pdf_creation_date(cc_coa_pdf)
    assert isinstance(pd.to_datetime(creation_date), datetime)
    creation_date = parser.get_pdf_creation_date(tagleaf_coa_pdf)
    assert isinstance(pd.to_datetime(creation_date), datetime)

    # [✓] TEST: `identify_lims`.
    identified_lims = parser.identify_lims(cc_coa_pdf)
    assert identified_lims == 'Confident Cannabis'
    identified_lims = parser.identify_lims(tagleaf_coa_pdf)
    assert identified_lims == 'TagLeaf LIMS'

    # [✓] TEST: Parse a PDF.
    data = parser.parse_pdf(cc_coa_pdf)

    # [✓] TEST: Parse a URL.
    data = parser.parse_url(cc_coa_url)

    # [✓] TEST: Parse a list of CoA URLs.
    urls = [cc_coa_url, tagleaf_coa_url]
    data = parser.parse(urls)

    # [✓] TEST: Parse a list of CoA PDFs.
    files = [cc_coa_pdf, tagleaf_coa_pdf]
    data = parser.parse(files)

    # [✓] TEST: Parse all CoAs in a given directory.
    data = parser.parse(DATA_DIR)

    # [ ] TEST: Parse all CoAs in a zipped folder!
    zip_folder = '../../../tests/assets/coas/coas.zip'
    data = parser.parse(zip_folder)
    assert data is not None

    # [✓] TEST: Parse a custom CoA (accept an error for now).
    try:
        parser.parse('https://cannlytics.page.link/partial-equilibrium-notes')
    except NotImplementedError:
        pass

    # [✓] TEST: Standardize CoA data.
    coa = 'https://lims.tagleaf.com/coa_/F6LHqs9rk9'
    # coa = '../../../.datasets/coas/Flore COA/Flore Brand/220121OgreInfused.pdf'
    data = parser.parse(coa)

    # [✓] TEST: Standardize CoA data dictionary.
    clean_data = parser.standardize(data[0])

    # [✓] TEST: Standardize CoA data list of dictionaries.
    clean_data_list = parser.standardize(data)

    # [✓] TEST: Standardize CoA data DataFrame.
    dataframe = pd.DataFrame(data)
    details_dataframe = parser.standardize(dataframe)
    results_dataframe = parser.standardize(
        dataframe,
        how='long'
    )
    values_dataframe = parser.standardize(
        dataframe,
        how='wide',
        details_data=details_dataframe,
        results_data=results_dataframe,
    )

    # [✓] TEST: Save CoA data from DataFrame.
    parser.save(dataframe, '../../../tests/assets/coas/test-coas.xlsx')

    # [✓] TEST: Save CoA data from list of dictionaries.
    parser.save(data, '../../../tests/assets/coas/test-coas.xlsx')

    # [✓] TEST: Save CoA data from dictionary.
    parser.save(data[0], '../../../tests/assets/coas/test-coas.xlsx')

    # [✓] TEST: Pass a PDF through OCR.
    doc = '../../../tests/assets/coas/210000068-Cloud-Cake-1g.pdf'
    temp_path = '../../../tests/assets/coas/tmp'
    temp_file = '../../../tests/assets/coas/tmp/ocr-coa.pdf'
    parser.pdf_ocr(doc, temp_file, temp_path=temp_path)

    # [✓] TEST: Parse a COA using OCR using `parse_pdf`.
    doc = '../../../tests/assets/coas/210000068-Cloud-Cake-1g.pdf'
    temp_path = '../../../tests/assets/tmp'
    data = parser.parse_pdf(doc, temp_path=temp_path)

    # [✓] TEST: Parse a COA using OCR using `parse`.
    doc = '../../../tests/assets/coas/210000068-Cloud-Cake-1g.pdf'
    temp_path = '../../../tests/assets/coas/tmp'
    data = parser.parse_pdf(doc, temp_path=temp_path)

    # [ ] TEST: Scan a QR code in an image with the `scan` method.
    parser = CoADoc()
    file_path = '../../../tests/assets/qr-code/steep-hill-qr-code.jpg'
    data = parser.scan(file_path)
    assert data.startswith('https')

    # [✓] TEST: Close the parser.
    parser.quit()
    print('All CoADoc tests finished.')



"""
Geographic Information Systems (GIS) Data | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 7/3/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    This script contains useful GIS functions.

"""
# Standard imports.
from datetime import datetime
from time import sleep
from typing import Any, List, Optional, Tuple

# External imports.
from dotenv import dotenv_values
from fredapi import Fred
from googlemaps import Client, places
import zipcodes

# Internal imports.
from cannlytics.firebase import initialize_firebase, get_document


def get_google_maps_api_key() -> str:
    """Get Google Maps API key.
    Returns:
        (str): Returns the Google Maps API key stored
            in a local .env file, Google Secret Manager, or Firestore.
    """
    import os
    from dotenv import load_dotenv
    
    # Try to get the key from .env file.
    load_dotenv()
    google_maps_api_key = os.getenv('GOOGLE_MAPS_API_KEY')
    if google_maps_api_key:
        return google_maps_api_key

    # Try to get the key from Google Secret Manager.
    try:
        from cannlytics.firebase import access_secret_version
        import google.auth
        _, project_id = google.auth.default()
        google_maps_api_key = access_secret_version(
            project_id=project_id,
            secret_id='GOOGLE_MAPS_API_KEY',
            version_id='latest',
        )
        if google_maps_api_key:
            return google_maps_api_key
    except Exception as e:
        print(f"Failed to get Google Maps API key from Secret Manager: {e}")

    # Try to get the key from Firestore.
    try:
        database = initialize_firebase()
        data = get_document('admin/google', database=database)
        if data and 'google_maps_api_key' in data:
            return data['google_maps_api_key']
    except Exception as e:
        print(f"Failed to get Google Maps API key from Firestore: {e}")

    # Raise an exception if the key is not found.
    raise Exception("Failed to get Google Maps API key")


def get_state_data(
        state: str,
        code: str,
        fred_api_key: Optional[str] = None,
        district: Optional[str] = '',
        obs_start: Optional[Any] = None,
        obs_end: Optional[Any] = None,
    ) -> dict:
    """Get a given state's data from the Fed Fred API, given a data code.
    Args:
        state (str): The state abbreviation for the state to retrieve data.
        code (str): The FRED code for the data, for example "POP".
        fred_api_key (str): A Fed FRED API key. You can sign up for a free API key at
            http://research.stlouisfed.org/fred2/. You can also pass `None`
            and set the environment variable 'FRED_API_KEY' to the value of
            your API key.
    Returns:
        (dict): Returns a dictionary with population values and source.
    """
    fred = Fred(api_key=fred_api_key)
    code = f'{state.upper()}{code.upper()}{district.upper()}'
    series = fred.get_series(code, obs_start, obs_end)
    if len(series) == 1:
        return series[0]
    return series


def get_state_population(
        state: str,
        fred_api_key: Optional[str] = None,
        district: Optional[str] = '',
        obs_start: Optional[Any] = None,
        obs_end: Optional[Any] = None,
        multiplier: Optional[float] = 1000.0,
    ) -> dict:
    """Get a given state's latest population from the Fed Fred API,
    getting the number in 1000's and returning the absolute value.
    Args:
        state (str): The state abbreviation for the state to retrieve
            population data. The abbreviation can be upper or lower case.
        fred_api_key (str): A Fed FRED API key. You can sign up for a free API key at
            http://research.stlouisfed.org/fred2/. You can also pass `None`
            and set the environment variable 'FRED_API_KEY' to the value of
            your API key.
    Returns:
        (dict): Returns a dictionary with population values and source.
    """
    pops = []
    fred = Fred(api_key=fred_api_key)
    code = f'{state.upper()}POP{district.upper()}'
    series = fred.get_series(code, obs_start, obs_end)
    for index, value in series.iteritems():
        real_pop = int(value * multiplier)
        pops.append({
            'population': real_pop,
            'population_formatted': f'{real_pop:,}',
            'population_source_code': code,
            'population_source': f'https://fred.stlouisfed.org/series/{code}',
            'population_at': index.isoformat()[:10],
        })
    if len(pops) == 1:
        return pops[0]
    return pops


def geocode_addresses(
        data,
        api_key: Optional[str] = None,
        pause: Optional[float] = 0.0,
        address_field: Optional[str] = '',
    ) -> Any:
    """Geocode addresses in a dataframe.
    Args:
        data (DataFrame): A DataFrame containing the addresses to geocode.
        api_key (str): A Google Maps API key.
        pause (float): An optional pause to wait between requests, 0.0 by default.
        address_field (str): An optional field to specify the address field,
            otherwise assumes the DataFrame has `street`, `city`, `state`,
            and `zip_code` columns.
    Returns:
        (DataFrame): Returns the DataFrame with `latitude`, `longitude`,
        and `formatted_address` columns.
    """
    if api_key is None:
        api_key = get_google_maps_api_key()
    gmaps = Client(key=api_key)
    for index, item in data.iterrows():
        if index and pause:
            sleep(pause)
        if address_field:
            address = item[address_field]
        else:
            address = f'{item.street}, {item.city}, {item.state} {item.zip_code}'
        geocode_result = gmaps.geocode(address)
        if geocode_result:
            data.at[index, 'formatted_address'] = geocode_result[0]['formatted_address']
            location = geocode_result[0]['geometry']['location']
            data.at[index, 'latitude'] = location['lat']
            data.at[index, 'longitude'] = location['lng']
            for info in geocode_result[0]['address_components']:
                key = info['types'][0]
                if key == 'administrative_area_level_1':
                    data.at[index, 'state'] = info['short_name']
                    data.at[index, 'state_name'] = info['long_name']
                if key == 'administrative_area_level_2':
                    data.at[index, 'county'] = info['long_name']
    return data


def search_for_address(
        query: str,
        api_key: Optional[str] = None,
        fields: Optional[List[str]] = None,
    ) -> dict:
    """Search for the address of a given name.
    Args:
        query (str): The text to use to search for an address.
        api_key (): Optional, None by default.
        fields (list): Optional, `formatted_address` is included by default.
    Returns:
        (list): A list of potential results.
    Raises:
        (IndexError): Raises an `IndexError` if no candidate is found.
    """
    if api_key is None:
        api_key = get_google_maps_api_key()
    if fields is None:
        fields = [
            'formatted_address',
            'geometry/location/lat',
            'geometry/location/lng',
        ]
    gmaps = Client(key=api_key)
    search = places.find_place(gmaps, query, 'textquery')
    place_id = search['candidates'][0]['place_id']
    place = places.place(gmaps, place_id, fields=fields)
    result = place['result']
    candidate = {}
    if result.get('geometry'):
        location = result['geometry']['location']
        candidate['latitude'] = location['lat']
        candidate['longitude'] = location['lng']
        del result['geometry']
    if result.get('formatted_address'):
        formatted_address = result['formatted_address']
        parts = formatted_address.split(',')
        if len(parts) == 4:
            candidate['street'] = parts[0]
            candidate['city'] = parts[1].strip()
            candidate['state'], candidate['zipcode'] = tuple(parts[2].strip().split(' '))
        else:
            candidate['city'] = parts[0].strip()
            candidate['state'], candidate['zipcode'] = tuple(parts[1].strip().split(' '))
        try:
            candidate['county'] = zipcodes.matching(candidate['zipcode'])[0]['county']
        except:
            candidate['county'] = ''
    return {**result, **candidate}


def get_transfer_distance(
        api_key,
        start,
        end,
        mode='driving',
) -> Tuple[int, int]:
    """Get the distance and duration of a transfer.
    Args:
        client (Client): A googlemaps API client.
        start (string): The starting point, either lat,long as a string or an address.
        end (string): The ending point, either lat,long as a string or an address.
        mode (string): The transportation method, driving by default.
    Returns:
        (int, int): Returns a tuple of the distance in kilometers and the
            duration in seconds.
    """
    client = Client(key=api_key)
    driving_distances = client.distance_matrix(start, end, mode=mode)
    elements = driving_distances['rows'][0]['elements'][0]
    km = elements['distance']['value']
    duration = elements['duration']['value']
    return km, duration


def get_transfer_route(
        api_key,
        start,
        end,
        departure_time=None,
        mode='driving',
) -> str:
    """Get the route of a transfer.
    Args:
        client (Client): A googlemaps API client.
        start (string): The starting point, either lat,long as a string or an address.
        end (string): The ending point, either lat,long as a string or an address.
        departure_time (datetime): The time of departure, defaults to now (optional).
        mode (string): The transportation method, driving by default (optional).
    Returns:
        (str): Returns the route as a polyline string.
    """
    client = Client(key=api_key)
    if departure_time is None:
        departure_time = datetime.now()
    driving_directions = client.directions(
        start,
        end,
        mode=mode,
        departure_time=departure_time
    )
    m = driving_directions[0]['legs'][0]['distance']['value']
    min = driving_directions[0]['legs'][0]['duration']['value']
    polyline = driving_directions[0]['overview_polyline']['points']
    return m, min, polyline


def initialize_googlemaps(env_file: Optional[str] = './.env') -> Any:
    """Initialize the Google Maps client.
    Args:
        env_file (str): A file path to a .env file with a `GOOGLE_MAPS_API_KEY`
            environment variable.
    Returns:
        (Client): A googlemaps API client.
    """
    config = dotenv_values(env_file)
    google_maps_api_key = config['GOOGLE_MAPS_API_KEY']
    return Client(key=google_maps_api_key)



"""
Contact Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Contact data model.
"""
# Standard imports.
from dataclasses import dataclass, field
from typing import List

# Internal imports.
from .base import Model


@dataclass
class Contact(Model):
    """Contacts in the case of non-laboratory users, are
    other organizations that you work with. Your contacts exist as
    organizations in the decentralized community, a network of
    federated servers that communicate with each other."""
    _collection = 'organizations/%s/contacts'
    additional_fields: List = field(default_factory=list)
    name: str = ''
    email: str = ''
    phone_number: str = ''
    phone_number_formatted: str = ''
    address_formatted: str = ''
    street: str = ''
    city: str = ''
    county: str = ''
    state: str = ''
    zip_code: str = ''
    latitude: float = 0.0
    longitude: float = 0.0
    people: List = field(default_factory=list)



"""
Cannlytics CoA Data Initialization | Cannlytics
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 7/21/2022
Updated: 8/30/2022
"""
try:
    from .coas import CoADoc
    __all__ = [
        CoADoc,
    ]
except ImportError:
    pass



"""
Metrc Constants | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 12/10/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

This module contains Metrc variables that are constant.
"""

# The number of minutes in the past to check the Metrc API
# when creating and updating objects and returning observations.
DEFAULT_HISTORY = 5

additive_types = [
    'Fertilizer',
    'Pesticide',
    'Other',
]

adjustment_reasons = [
    {
        'Name': 'API Related Error',
        'RequiresNote': True,
    },
    {
        'Name': 'Drying',
        'RequiresNote': True,
    },
    {
        'Name': 'Mandatory State Destruction',
        'RequiresNote': True,
    },
    {
        'Name': 'Theft',
        'RequiresNote': True,
    },
    {
        'Name': 'Typing/Entry Error',
        'RequiresNote': True,
    }
]

analyses = [
    {
        'Id': 1,
        'Name': 'THC',
        'RequiresTestResult': False,
        'InformationalOnly': False,
        'AlwaysPasses': False,
        'LabTestResultMode': 0,
        'LabTestResultMinimum': None,
        'LabTestResultMaximum': None,
        'DependencyMode': 0
    },
    {
        'Id': 2,
        'Name': 'THCa',
        'RequiresTestResult': False,
        'InformationalOnly': False,
        'AlwaysPasses': False,
        'LabTestResultMode': 0,
        'LabTestResultMinimum': None,
        'LabTestResultMaximum': None,
        'DependencyMode': 0
    },
    {
        'Id': 3,
        'Name': 'CBD',
        'RequiresTestResult': False,
        'InformationalOnly': False,
        'AlwaysPasses': False,
        'LabTestResultMode': 0,
        'LabTestResultMinimum': None,
        'LabTestResultMaximum': None,
        'DependencyMode': 0
    },
    {
        'Id': 4,
        'Name': 'CBDa',
        'RequiresTestResult': False,
        'InformationalOnly': False,
        'AlwaysPasses': False,
        'LabTestResultMode': 0,
        'LabTestResultMinimum': None,
        'LabTestResultMaximum': None,
        'DependencyMode': 0
    },
    {
        'Id': 5,
        'Name': 'Pesticides',
        'RequiresTestResult': False,
        'InformationalOnly': False,
        'AlwaysPasses': False,
        'LabTestResultMode': 0,
        'LabTestResultMinimum': None,
        'LabTestResultMaximum': None,
        'DependencyMode': 'RequiresOne'
    }
]

batch_types = [
    'Seed',
    'Clone',
]

categories = [
    'Flower & Buds',
    'Immature Plants',
    'Concentrate (Non-Solvent Based) (Count-Volume)',
    'Concentrate (Non-Solvent Based) (Count-Weight)',
    'Concentrate (Weight Based)',
    'Edibles (Count-Volume)',
    'Edibles (Count-Weight)',
    'Extracts (Solvent Based) (Count-Volume)',
    'Extracts (Solvent Based) (Count-Weight)',
    'Kief',
    'Mature Plants',
    'Metered Dose Nasal Spray Products',
    'MMJ Waste',
    'Pre-Roll (Flower Only)',
    'Pre-Roll (Infused)',
    'Pressurized Metered Dose Inhaler Products',
    'Rectal/Vaginal Administration Products (Count-Volume)',
    'Rectal/Vaginal Administration Products (Count-Weight)',
    'Seeds',
    'Shake/Trim',
    'Shake/Trim (by Strain)',
    'Tinctures (Count-Volume)',
    'Tinctures (Count-Weight)',
    'Topicals (Count-Volume)',
    'Topicals (Count-Weight)',
    'Transdermal Patches',
    'Vape Cartridges',
    'Whole Wet Plant',
]

customer_types = [
    'Consumer',
    'Patient',
    'Caregiver',
    'ExternalPatient',
]

growth_phases = [
    'Young',
    'Vegetative',
    'Flowering',
]

item_types = [
    {
        'Name': 'Buds',
        'ProductCategoryType': 'Buds',
        'QuantityType': 'WeightBased',
        'RequiresStrain': True,
        'RequiresItemBrand': False,
        'RequiresAdministrationMethod': False,
        'RequiresUnitCbdPercent': False,
        'RequiresUnitCbdContent': False,
        'RequiresUnitCbdContentDose': False,
        'RequiresUnitThcPercent': False,
        'RequiresUnitThcContent': False,
        'RequiresUnitThcContentDose': False,
        'RequiresUnitVolume': False,
        'RequiresUnitWeight': False,
        'RequiresServingSize': False,
        'RequiresSupplyDurationDays': False,
        'RequiresNumberOfDoses': False,
        'RequiresPublicIngredients': False,
        'RequiresDescription': False,
        'RequiresProductPhotos': 0,
        'RequiresLabelPhotos': 0,
        'RequiresPackagingPhotos': 0,
        'CanContainSeeds': True,
        'CanBeRemediated': True
    },
    {
        'Name': 'Immature Plants',
        'ProductCategoryType': 'Plants',
        'QuantityType': 'CountBased',
        'RequiresStrain': True,
        'RequiresItemBrand': False,
        'RequiresAdministrationMethod': False,
        'RequiresUnitCbdPercent': False,
        'RequiresUnitCbdContent': False,
        'RequiresUnitCbdContentDose': False,
        'RequiresUnitThcPercent': False,
        'RequiresUnitThcContent': False,
        'RequiresUnitThcContentDose': False,
        'RequiresUnitVolume': False,
        'RequiresUnitWeight': False,
        'RequiresServingSize': False,
        'RequiresSupplyDurationDays': False,
        'RequiresNumberOfDoses': False,
        'RequiresPublicIngredients': False,
        'RequiresDescription': False,
        'RequiresProductPhotos': 0,
        'RequiresLabelPhotos': 0,
        'RequiresPackagingPhotos': 0,
        'CanContainSeeds': True,
        'CanBeRemediated': False
    },
    {
        'Name': 'Infused',
        'ProductCategoryType': 'InfusedEdible',
        'QuantityType': 'CountBased',
        'RequiresStrain': False,
        'RequiresItemBrand': False,
        'RequiresAdministrationMethod': False,
        'RequiresUnitCbdPercent': False,
        'RequiresUnitCbdContent': False,
        'RequiresUnitCbdContentDose': False,
        'RequiresUnitThcPercent': False,
        'RequiresUnitThcContent': True,
        'RequiresUnitThcContentDose': False,
        'RequiresUnitVolume': False,
        'RequiresUnitWeight': True,
        'RequiresServingSize': False,
        'RequiresSupplyDurationDays': False,
        'RequiresNumberOfDoses': False,
        'RequiresPublicIngredients': False,
        'RequiresDescription': False,
        'RequiresProductPhotos': 0,
        'RequiresLabelPhotos': 0,
        'RequiresPackagingPhotos': 0,
        'CanContainSeeds': False,
        'CanBeRemediated': True
    },
    {
        'Name': 'Infused Liquid',
        'ProductCategoryType': 'InfusedEdible',
        'QuantityType': 'CountBased',
        'RequiresStrain': False,
        'RequiresItemBrand': False,
        'RequiresAdministrationMethod': False,
        'RequiresUnitCbdPercent': False,
        'RequiresUnitCbdContent': False,
        'RequiresUnitCbdContentDose': False,
        'RequiresUnitThcPercent': False,
        'RequiresUnitThcContent': True,
        'RequiresUnitThcContentDose': False,
        'RequiresUnitVolume': True,
        'RequiresUnitWeight': False,
        'RequiresServingSize': False,
        'RequiresSupplyDurationDays': False,
        'RequiresNumberOfDoses': False,
        'RequiresPublicIngredients': False,
        'RequiresDescription': False,
        'RequiresProductPhotos': 0,
        'RequiresLabelPhotos': 0,
        'RequiresPackagingPhotos': 0,
        'CanContainSeeds': False,
        'CanBeRemediated': True
    }
]

location_types = [
    {
        'Id': 1,
        'Name': 'Default',
        'ForPlantBatches': True,
        'ForPlants': True,
        'ForHarvests': True,
        'ForPackages': True,
    },
    {
        'Id': 2,
        'Name': 'Planting',
        'ForPlantBatches': True,
        'ForPlants': True,
        'ForHarvests': False,
        'ForPackages': False,
    },
    {
        'Id': 3,
        'Name': 'Packing',
        'ForPlantBatches': False,
        'ForPlants': False,
        'ForHarvests': False,
        'ForPackages': True,
    },
]

package_types = [
    'Product',
    'ImmaturePlant',
    'VegetativePlant',
    'PlantWaste',
    'HarvestWaste',
]

parameters = {
    'license_number': 'licenseNumber',
    'start': 'lastModifiedStart',
    'end': 'lastModifiedEnd',
    'sales_start': 'salesDateStart',
    'sales_end': 'salesDateEnd',
    'package_id': 'packageId',
    'from_mother': 'isFromMotherPlant',
    'source': 'source',
}

# TODO: Rejection reasons vary by state...
# rejection_types = {}

test_statuses = [
    'NotSubmitted',
    'SubmittedForTesting',
    'TestFailed',
    'TestPassed',
    'TestingInProgress',
    'AwaitingConfirmation',
    'RetestFailed',
    'RetestPassed',
    'Remediated',
    'SelectedForRandomTesting',
    'NotRequired',
    'ProcessValidated',
]

transfer_statuses = [
    'Shipped',
    'Rejected',
    'Accepted',
    'Returned',
]

# TODO: Transfer types vary by state...
transfer_types = [
    {
        'Name': 'Affiliated Transfer',
        'ForLicensedShipments': True,
        'ForExternalIncomingShipments': False,
        'ForExternalOutgoingShipments': False,
        'RequiresDestinationGrossWeight': False,
        'RequiresPackagesGrossWeight': False,
    },
    {
        'Name': 'Beginning Inventory Transfer',
        'ForLicensedShipments': False,
        'ForExternalIncomingShipments': True,
        'ForExternalOutgoingShipments': False,
        'RequiresDestinationGrossWeight': False,
        'RequiresPackagesGrossWeight': False,
    },
    {
        'Name': 'Lab Sample Transfer',
        'ForLicensedShipments': True,
        'ForExternalIncomingShipments': False,
        'ForExternalOutgoingShipments': False,
        'RequiresDestinationGrossWeight': False,
        'RequiresPackagesGrossWeight': False,
    },
    {
        'Name': 'Unaffiliated (Wholesale) Transfer',
        'ForLicensedShipments': True,
        'ForExternalIncomingShipments': False,
        'ForExternalOutgoingShipments': False,
        'RequiresDestinationGrossWeight': False,
        'RequiresPackagesGrossWeight': False,
    },
    {
        'Name': 'Waste Disposal',
        'ForLicensedShipments': True,
        'ForExternalIncomingShipments': False,
        'ForExternalOutgoingShipments': False,
        'RequiresDestinationGrossWeight': False,
        'RequiresPackagesGrossWeight': False,
    }
]

units = [
    {
        'QuantityType': 'CountBased',
        'Name': 'Each',
        'Abbreviation': 'ea',
    },
    {
        'QuantityType': 'WeightBased',
        'Name': 'Ounces',
        'Abbreviation': 'oz',
    },
    {
        'QuantityType': 'WeightBased',
        'Name': 'Pounds',
        'Abbreviation': 'lb',
    },
    {
        'QuantityType': 'WeightBased',
        'Name': 'Grams',
        'Abbreviation': 'g',
    },
    {
        'QuantityType': 'WeightBased',
        'Name': 'Milligrams',
        'Abbreviation': 'mg',
    },
    {
        'QuantityType': 'WeightBased',
        'Name': 'Kilograms',
        'Abbreviation': 'kg',
    },
]

waste_methods = [
    {'Name': 'Grinder'},
    {'Name': 'Compost'},
]

waste_reasons = [
    {
        'Name': 'Disease/Infestation',
        'RequiresNote': True,
    },
    {
        'Name': 'Mother Plant Destruction',
        'RequiresNote': True,
    },
    {
        'Name': 'Trimming/Pruning',
        'RequiresNote': False,
    }
]

# TODO: Waste types vary by state...
# harvest_waste_types = [
#     'Plant Material',
#     'Fibrous',
#     'Root Ball',
# ]
# harvest_waste_types = [
#     'MMJ Waste',
#     'Waste',
# ]
waste_types = {
    'ca': ['Plant Material'],
}



"""
Sample Management | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 6/23/2021  
Updated: 6/23/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Manage your laboratory samples.
"""

def get_sample_pictures():
    """Get sample pictures taken and stored on a lab computer."""
    # pics = []
    # return pics
    return NotImplementedError


if __name__ == '__main__':
    
    # TODO: Test get sample pictures.
    get_sample_pictures()




"""
Parse Anresco Laboratories CoAs
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 8/2/2022
Updated: 9/7/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse Anresco Laboratories CoA PDFs and URLs.

Data Points:

    - analyses
    - {analysis}_method
    ✓ {analysis}_status
    ✓ batch_number
    ✓ date_collected
    ✓ date_received
    ✓ date_tested
    - images
    ✓ lab_results_url
    - metrc_source_id
    ✓ project_id
    ✓ producer
    ✓ producer_address
    ✓ product_name
    ✓ product_type
    - results
    ✓ sample_weight
    ✓ status
    - total_cannabinoids
    - total_thc
    - total_cbd
    - total_terpenes
    ✓ sample_id (generated)
    ✓ lab_id
    ✓ lab
    ✓ lab_image_url
    ✓ lab_license_number
    ✓ lab_address
    ✓ lab_city (augmented)
    ✓ lab_county (augmented)
    ✓ lab_state (augmented)
    ✓ lab_zipcode (augmented)
    ✓ lab_latitude (augmented)
    ✓ lab_longitude (augmented)

"""
# Standard imports.
from ast import literal_eval
from datetime import datetime
import json
import re
from typing import Any

# External imports.
# from bs4 import BeautifulSoup
import pandas as pd
import pdfplumber
# import requests

# Internal imports.
from cannlytics.data.data import create_sample_id, find_first_value
from cannlytics.utils.constants import ANALYTES, STANDARD_FIELDS
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    # split_list,
    # strip_whitespace,
)

# External imports.
import pandas as pd
# from selenium import webdriver
# from selenium.webdriver.chrome.options import Options
# from selenium.webdriver.common.by import By
# from selenium.webdriver.chrome.service import Service
# from selenium.common.exceptions import (
#     ElementNotInteractableException,
#     TimeoutException,
# )
# from selenium.webdriver.support import expected_conditions as EC
# from selenium.webdriver.support.ui import WebDriverWait
# try:
#     import chromedriver_binary  # Adds chromedriver binary to path.
# except ImportError:
#     print('Proceeding assuming that you have ChromeDriver in your path.')


# It is assumed that the lab has the following details.
ANRESCO =  {
    'coa_algorithm': 'anresco.py',
    'coa_algorithm_entry_point': 'parse_anresco_coa',
    'lims': 'Anresco Laboratories',
    'lab': 'Anresco Laboratories',
    'lab_image_url': 'https://anresco.com/wp-content/uploads/thegem-logos/logo_2258c6b0a6f4e626a22edd2f64dc588b_2x.png',
    'lab_website': 'www.anresco.com',
    'lab_address': '1375 Van Dyke Ave, San Francisco, CA 94124',
    'lab_street': '1375 Van Dyke Ave',
    'lab_city': 'San Francisco',
    'lab_state': 'CA',
    'lab_zipcode': '94124',
    'lab_latitude': 37.726610,
    'lab_longitude': -122.388450,
    'public': True,
}

# It is assumed that the CoA has the following parameters.
ANRESCO_COA = {
    'coa_distributor_area': '(200, 60, 380, 130)',
    'coa_lab_area': '(0, 60, 200, 130)',
    'coa_producer_area': '(380, 60, 580, 130)',
    'coa_sample_details_area': [
        '(200, 130, 385, 310)',
        '(385, 130, 590, 310)',
    ],
    'coa_drop_columns': [
        'sample_weight_to',
        'sample_information',
        'test_summary',
        'harvest',
    ],
}


def parse_anresco_pdf(parser, doc: Any, **kwargs) -> Any:
    """Parse a Anresco Laboratories CoA PDF.
    Args:
        parser (CoADoc): A CoADoc parsing client.
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = report.stream.name.replace('\\', '/').split('/')[-1]
    front_page = report.pages[0]

    # Get the QR code from the last page.
    obs['lab_results_url'] = parser.find_pdf_qr_code_url(report, page_index=-1)

    # Get the standard analyses, analytes, and fields.
    coa_parameters = ANRESCO_COA

    # Get the sample details based on page area.
    sample_details_area = coa_parameters['coa_sample_details_area']
    if isinstance(sample_details_area, str):
        sample_details_area = [sample_details_area]
    for area in sample_details_area:
        crop = front_page.within_bbox(literal_eval(area))
        lines = crop.extract_text().split('\n')
        for line in lines:
            parts = line.replace('\xa0', ' ').split(':')
            key = snake_case(parts[0])
            key = STANDARD_FIELDS.get(key, key)
            if not key:
                continue
            value = parts[-1].replace('✔', '')
            obs[key] = value.strip()

    # Hot-fix.
    drop_columns = coa_parameters['coa_drop_columns']
    for column in drop_columns:
        try:
            del obs[column]
        except:
            pass

    # Get the lab data based on page area.
    area = coa_parameters['coa_lab_area']
    crop = front_page.within_bbox(literal_eval(area))
    lines = crop.extract_text().split('\n')[1:]
    obs['lab'] = lines[0]
    obs['lab_address'] = ', '.join([lines[1], lines[2]])
    obs['lab_license_number'] = lines[-1]

    # Get the producer data based on page area.
    area = coa_parameters['coa_producer_area']
    crop = front_page.within_bbox(literal_eval(area))
    lines = crop.extract_text().split('\n')[1:]
    obs['producer'] = lines[0]
    obs['producer_address'] = ', '.join([lines[1], lines[2]])
    obs['producer_license_number'] = lines[-1]

    # Get the distributor data based on page area.
    area = coa_parameters['coa_distributor_area']
    crop = front_page.within_bbox(literal_eval(area))
    lines = crop.extract_text().split('\n')[1:]
    obs['distributor'] = lines[0]
    obs['distributor_address'] = ' '.join([lines[1], lines[2]])
    obs['distributor_license_number'] = lines[-1]

    # Future work: Get all GIS data for `lab`, `producer`, and `distributor`.

    # Get the result data.
    # TODO: Get the `analyses` and `{analysis}_status`.
    # TODO: Add `analysis` to results.
    analyses = []
    # analysis = None
    columns = []
    results = []
    # units = None
    for page in report.pages:

        # Iterate over the lines of text on the page.
        lines = page.extract_text().split('www.anresco.com')[0].split('\n')
        for line in lines:

            # TODO: Get the analysis.

            # Get the totals.
            if line.startswith('Total'):

                first_value = find_first_value(line)
                name = line[:first_value].strip()
                key = snake_case(name)
                key = STANDARD_FIELDS.get(key, key)
                values = line[first_value:].strip().split(' ')
                values = [x for x in values if x]
                obs[key] = convert_to_numeric(values[-1])

            # Get the columns.
            # Note: Hot-fix for cannabinoids.
            elif line.startswith('Analyte') or line.startswith('Cannabinoid mg/g'):

                # Try to get the units.
                # FIXME: Determine units another way?
                # if '(' in line:
                #     units = line.split('(')[-1].split(')')[0]
                # elif line.startswith('Cannabinoid'):
                #     units = 'percent'
                # else:
                #     units = None

                # Get the standard columns.
                text = re.sub('[\(\[].*?[\)\]]', '', line)
                columns = [
                    STANDARD_FIELDS[x] for x in text.split(' ') if x
                ]

            # Get the analysis method.
            # FIXME: Not all analyses have a method and instrument.
            elif 'Method:' in line:
                continue

            elif 'Instrument:' in line:
                continue

            # Skip extraneous rows.
            elif not line or not columns or '✔' in line:
                continue

            # Break at the end of the CoA.
            elif line.startswith('(-) = Not Tested'):
                break

            # Extract the results!
            else:

                # Find the first value.
                first_value = find_first_value(line)
                if not first_value:
                    # FIXME: Microbes have text values.
                    continue

                # FIXME: Hot-fix to collect STEC result.
                elif line.startswith('STEC'):
                    values = line.split(' ')
                    results.append({
                        'analysis': 'microbes',
                        'name': values[0],
                        'key': snake_case(name),
                        'method': ' '.join(parts[1:-3]).strip(),
                        'value': values[-3],
                        'status': values[-1],
                    })
                    continue

                # FIXME: Hot-fix for foreign matter units.
                line = line.replace('1 per 3g', '1/3g')

                # Parse the result values.
                name = line[:first_value].strip()
                analyte = snake_case(name)
                analyte = ANALYTES.get(analyte, analyte)
                values = line[first_value:].strip().split(' ')
                values = [x for x in values if x]
                result = {
                    'analysis': None, # FIXME: <-- Add this data point!
                    'key': analyte,
                    'name': name,
                    'units': None, # FIXME: <-- Add this data point!
                }
                try:
                    for i, value in enumerate(values):
                        key = columns[i + 1]
                        result[key] = convert_to_numeric(value)
                except IndexError:
                    continue 
                results.append(result)

    # Hot-fix: Collect the microbes!
    columns = []
    area = (0, 600, 595, 680)
    crop = front_page.within_bbox(area)
    table = crop.extract_table({
        'vertical_strategy': 'text',
    })
    for row in table:
        name = ' '.join(row[0:2]).strip()
        key = snake_case(name)
        method = ' '.join(row[2:-2]).strip()
        results.append({
            'analysis': 'microbes',
            'key': key,
            'name': name,
            'value': row[-2],
            'status': row[-1],
            'method': method,
        })
    
    # Hot-fix: Remove any results without values.
    results = [x for x in results if 'value' in x]

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Finish data collection with a freshly minted sample ID.
    obs['analyses'] = analyses
    obs['results'] = results
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs['producer'],
    )
    obs['coa_parsed_at'] = datetime.now().isoformat()
    return {**ANRESCO, **obs}


#--------------------------------------------
# TODO: Parse a Anresco Laboratories CoA URL.
#--------------------------------------------

def parse_anresco_url(parser, doc: Any, **kwargs) -> Any:
    """Parse a Anresco Laboratories CoA URL.
    Args:
        parser (CoADoc): A CoADoc parsing client.
        doc (str or PDF): A PDF file path or pdfplumber PDF or URL.
    Returns:
        (dict): The sample data.
    """
    # FIXME: Allow the user to pass a URL directly
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
    else:
        report = doc
    front_page = report.pages[0]

    # Get the QR code from the last page.
    obs = {}
    obs['lab_results_url'] = parser.find_pdf_qr_code_url(report, page_index=-1)

    # TODO: Parse the data from the Anresco portal.

    raise NotImplementedError


#--------------------------
# Under development
#--------------------------

# DEV: Default parameters
# max_delay = 7

# # # Get the HTML!
# # headers = DEFAULT_HEADERS
# # response = requests.get(url, headers=headers)
# # soup = BeautifulSoup(response.content, 'html.parser')
# service = Service()
# options = Options()
# options.add_argument('--window-size=1920,1200')
# # Uncomment for dev:
# # parser.options.headless = False
# # Uncomment for production!!!
# # options.add_argument('--headless')
# # options.add_argument('--disable-gpu')
# # options.add_argument('--no-sandbox')
# driver = webdriver.Chrome(
#     options=options,
#     service=service,
# )
# driver.get(url)

# # Wait for the page to load.
# try:
#     el = (By.CLASS_NAME, 'public-sample')
#     detect = EC.presence_of_element_located(el)
#     WebDriverWait(driver, max_delay).until(detect)
# except TimeoutException:
#     print('Failed to load page within %i seconds.' % max_delay)

# # TODO: Get the lab details.

# # TODO: Get the distributor details.

# # TODO: Get the producer details.

# # Get the image URL.
# el = driver.find_element_by_css_selector('[alt="Sample picture"]')
# image_url = el.get_attribute('src')
# filename = image_url.split('/')[-1]
# obs['images'] = [{'url': image_url, 'filename': filename}]

# # Get the sample details.
# button = driver.find_element(by=By.CLASS_NAME, value='ps-show-more-btn')
# button.click()
# standard_fields = ANRESCO['coa_fields']
# fields = driver.find_elements(by=By.CLASS_NAME, value='ps-sample-info-table-static')
# values = driver.find_elements(by=By.CLASS_NAME, value='ps-sample-info-table-variable')
# for i, value in enumerate(values):
#     key = fields[i].text
#     key = standard_fields.get(key, snake_case(key))
#     obs[key] = value.text

# # Open all of the results.
# els = driver.find_elements(by=By.CLASS_NAME, value='fa-chevron-down')
# for el in els:
#     el.click()

# # Iterate over the sections.
# methods = []
# sections = driver.find_elements(by=By.CLASS_NAME, value='ps-mobile-section-expanded')

# # DEV:
# section = sections[0]

# # TODO: Get the analysis.

# # TODO: Get the analysis status.

# # Get the method.
# el = section.find_element(by=By.CLASS_NAME, value='ps-section-description')
# method = el.text
# methods.append({'analysis': '', 'method': method})


# # FIXME: Get all of the results!


# # TODO: Get the columns and units.
# # ps-analyte-th


# # TODO: Get and parse each analyte row.
# # ps-analyte


# # TODO: Get the LOD and LOQ for each result.


# # TODO: Get the totals!
# # - total_cannabinoids
# # - total_thc
# # - total_cbd


# print({**ANRESCO, **obs})


def parse_anresco_coa(parser, doc: Any, **kwargs) -> Any:
    """Parse an Anresco Laboratories CoA PDF or URL."""
    # FIXME: Handle both PDFs and URLs.
    # # Read the PDF.
    # if isinstance(doc, str):
    #     report = pdfplumber.open(doc)
    # else:
    #     report = doc
    # front_page = report.pages[0]

    # # Get the QR code from the last page.
    # lab_results_url = parser.find_pdf_qr_code_url(report, page_index=-1)
    return parse_anresco_pdf(parser, doc, **kwargs)


#--------------------------
# Future work
#--------------------------

# Standardize analytes with NLP!


# === Tests ===
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc

    # Test parsing Anresco Laboratories CoAs.
    parser = CoADoc()
    doc = '../../../tests/assets/coas/Peanutbutter Breath.pdf'
    coa_url = 'https://portal.anresco.com/#/ps/68bcb967c0f2b39d'

    # [✓] TEST: Identify a given PDF or URL as an Anresco CoA.
    lab = parser.identify_lims(doc, lims={'Anresco Laboratories': ANRESCO})
    assert lab == 'Anresco Laboratories'

    # [✓] TEST: Parse an Anresco Laboratories CoA PDF.
    parser = CoADoc()
    data = parse_anresco_pdf(parser, doc)
    assert data is not None

    # [ ] TEST: Parse an Anresco Laboratories CoA URL.
    # parser = CoADoc()
    # data = parse_anresco_url(parser, doc)
    # assert data is not None



"""
Cannlytics Statistics Initialization | Cannlytics
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 5/31/2022
Updated: 5/31/2022
"""

from .stats import (
    calculate_model_statistics,
    estimate_discrete_model,
    get_stats_model,
    predict_stats_model,
    upload_stats_model,
)

__all__ = [
    'calculate_model_statistics',
    'estimate_discrete_model',
    'get_stats_model',
    'predict_stats_model',
    'upload_stats_model',
]
